unit KnobsConversions;

{

  (C) COPYRIGHT 2014 .. 2016 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Classes, System.Math, Vcl.Graphics,

  ComplexMath, KnobsUtils, Rationals;


const

  DELTA_LIMIT       = 1e-6;                          // Small value for equality comparisons                                      {  -120 dB}}
  SMALL_VALUE       = 1e-50;                         // Lowest 'tolerable' signal level in patches                                ( -1000 dB)
  NORMAL_LIMIT      = 1e-20;                         // Value below which the Normalize functions will truncate the value to zero {  -400 dB}
  MUTE_LIMIT        = 1e-7;                          // Value below which mute and knob signals are assumed to be off             (  -140 dB)
  EULER             = 2.718281828459045235360287471; // e .. as in Euler's constant
  TWO_PI            = PI * 2;                        // Two times Pi, for angle conversions
  TWO_PI_REC        = 1.0 / TWO_PI;                  // 1 divided by two times PI
  HALF_PI           = PI / 2;                        // Pi / 2      ; for fade tables and for S-envelopes
  SQRT2             = 1.4142136;                     // Sqrt( 2)
  SQRT2_REC         = 1 / SQRT2;                     // Reciprocal of SQRT_2 :: 0.7071067
  THIRD             = 1.259921;                      // 2 ^ (1/3)
  THIRD_REC         = 1 / THIRD;                     // reciprocal of THIRD :: 0.7937005
  ONE_THIRD         = 1.0 / 3.0;                     // just that 1/3 - 0.3333
  TWO_THIRDS        = 2.0 / 3.0;                     // just that 2/3 - 0.6667
  NOTE_SCALING      = 128.0;                         // Value used to scale note numberss to/from internal values
  NOTE_SCALING_REC  = 1 / NOTE_SCALING;              // Reciprocal of NOTE_SCALING
  MIDI_SCALING      = 128.0;                         // Value to scale MIDI values to/from internal signals
  MIDI_SCALING_REC  = 1 / MIDI_SCALING;              // Reciprocal of MIDI_SCALING
  CONTROL_DEC       = 8;                             // !! POWER OF TWO !! Decimation used for slow tick stuff ( control rate)
  ONE_TWELFTH       = 1 / 12.0;                      // Reciprocal of 12
  NO_NOTE           = 1.23456e7;                     // ** MAGIC NUMBER ** for sequencers to play no note
  MAX_DB_VAL        = 140.0;                         // Table scaling for dB lookup, range is [- DB_SCALE, + DB_SCALE]
  MAX_DB_VAL_REC    = 1.0/ MAX_DB_VAL;               // Reciprocal of MAX_DB_VAL
  MIN_DB_SIGNAL     = 1e-7;                          // Minimum signal value for which deciBels can be calculated
  MAX_DB_SIGNAL     = 1e7;                           // Maximum signal value for which deciBels can be calculated
  MIN_DB_SIGNAL_REC = 1.0 / MIN_DB_SIGNAL;           // Reciprocal of MIN_DB_SIGNAL;
  MAX_DB_SIGNAL_REC = 1.0 / MAX_DB_SIGNAL;           // Reciprocal of MAX_DB_SIGNAL;
  CENTS             = 100.0;                         // Scaling for cents lookup
  CENTS_REC         = 1.0 / CENTS;                   // reciprocal of CENTS
  FRACT_SCALING     = 1 shl 26;                      // To scale from TSignal to fractional values - needs to match SmallForth setting in fixpoint.s4
  FRACT_SCALING_REC = 1 / FRACT_SCALING;             // To scale from fractional values to TSignal
  KIBI              = 1024;                          // Binary multiplication factors
  MEBI              = KIBI * KIBI;
  GIBI              = KIBI * MEBI;
  MIN_IN_SECS       = 1.0 / 60.0;                    // Minutes in second
  UNITS_IN_CENT     = 1.0 / 256.0;
  UNITS_IN_DEGREE   = 1.0 / 360.0;
  NO_UNITS          = 'no_units';                    // For conveters not accepting any units


  ShortestPossibleEnvelopeTime =      1;             // In samples
  EnvelopeSmoothing            =     99;             // envout := ( EnvelopeSmooting * OldValue + NewValue) / ( 1 + EnvelopeSmoothing); - was 49
  EnvelopeSmoothingSlow        =     19;             // envout := ( EnvelopeSmooting * OldValue + NewValue) / For control rate envelope processing
  MinEnvTimeFast               =      0.0005;        // Minimum time for fast   rate envelopes in seconds
  MinEnvTimeMedium             =      0.05;          // Minimum time for medium rate envelopes in seconds
  MinEnvTimeSlow               =      5;             // Minimum time for slow   rate envelopes in seconds
  MaxEnvTimeFast               =     12;             // Maximum time for fast   rate envelopes in seconds
  MaxEnvTimeMedium             =   1200;             // Maximum time for medium rate envelopes in seconds
  MaxEnvTimeSlow               = 120000;             // Maximum time for slow   rate envelopes in seconds
  CompressorSmoothing          = 1.0 / 20.0;         // envelope smoothing for compressor - was 9 then was 19, now is a float with value 1/20

  {$IFDEF DEBUG}
    UNDEFINED = NAN;                                 // NAN for stability checks
    {$INLINE OFF}                                    // No inlining for debug code
  {$ELSE}
    UNDEFINED = 0.0;                                 // 0.0 for production code
    {$INLINE ON}                                     // Try to use inlining on production code
  {$ENDIF}


var

  ReferenceA         : TSignal;  // The value of the reference A to be used - default 440 Hz
  NotesPerOctave     : TSignal;  // The number of notes wanted in an octave - there only is notation for 12 tho ... A .. G#
  NotesPerOctaveRec  : TSignal;  // Reciprocal of NotesPerOctave
  MiddleNote         : TSignal;  // The middle note to be used - default 69 - this is a note number


  // -------------------------------------------------------------------------------------------------------------------
  // NOTE : the following five CAN NOT BE CHANGED at will, some relations need to be maintained
  // for this see **1 and please use the SetSystemRate function to change the System_Rate.
  // -------------------------------------------------------------------------------------------------------------------
  // These can not be moved to the implementation section as that will prevent inlining of
  // some functions. Also after changing these audio interfaces need to be re-initialized and lookuptables
  // for modules must be regenerated.

  System_Rate        : Integer;  // 44100.0;                     // samples / s  ( audio   rate)
  System_Rate_2      : TSignal;  // System_Rate * 2;             // samples / s
  Control_Rate       : TSignal;  // System_Rate / CONTROL_DEC;   // samples / s
  System_Rate_Rec    : TSignal;  // reciprocal of system rate    // s / sample
  System_Rate_2_Rec  : TSignal;  // reciprocal of system rate_2  // s / sample

  // END NOTE.
  // -------------------------------------------------------------------------------------------------------------------


type

  EKnobsConversions   = class( Exception);
  ENoDefaultConverter = class( EKnobsConversions);


  TSignalRate = (
    srControl,
    srAudio
  );


  TMidiSeptet = 0 .. 127;
  TMidiFilter = set of TMidiSeptet;
  TEnvRange   = ( erSlow, erMedium, erFast);


type

  TSignalFunction = function( const aValue: TSignal): TSignal;


  TSignalArrayWrapper = class
  public
    FSignals    : TSignalArray;
    FlastOutput : TSignal;
  private
    function    GetCount : Integer;
    function    GetSignal( anIndex: Integer): TSignal;
    procedure   SetSignal( anIndex: Integer; const aValue: TSignal);
  public
    constructor Create( aCount: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    procedure   Erase;
    procedure   AddSignal( const aValue: TSignal);
    procedure   SetSignals( aStartIndex: Integer; const aValue: TSignalArrayWrapper);
    function    InterpolateL( const anIndex: TSignal): TSignal;   // index [0, Count-1] - linear   interpolation
    function    InterpolateA( const anIndex: TSignal): TSignal;   // index [0, Count-1] - all-pass interpolation
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal; default;
  end;


  TSignalPairFifo = class
  private
    FSignals        : TSignalPairArray; // The data buffer
    FInp            : Integer;          // Current buffer insertion  point
    FOutp           : Integer;          // Current buffer extraction point
    FFreeCount      : Integer;          // Number of free data slots left
    FFillCount      : Integer;          // Number of filled data slots available
    FAllowOverflows : Boolean;          // True when overflows are allowed, pld data is sicarded then
    FOverflows      : Integer;          // Number of times Append  was called while the buffer had no free space
    FUnderflows     : Integer;          // Number of tiems GetData was called while the buffer had no items in it
    FLastValue      : TSignalPair;      // To repeatedly return the last pair when no samples are available
    FCountOverflows : Boolean;          // Set to true to keep count of overflows
  private
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
    function    GetSize      : Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    procedure   Append( const aLeft, aRight: TSignal);                                                         overload;
    procedure   Append( const aPair: TSignalPair);                                                             overload;
    function    GetData: TSignalPair;
    procedure   ClearOverUnderFlows;
  public
    property    Overflows      : Integer read GetOverflows;
    property    Underflows     : Integer read GetUnderflows;
    property    FillCount      : Integer read FFillCount;
    property    FreeCount      : Integer read FFreeCount;
    property    Size           : Integer read GetSize;
    property    AllowOverflows : Boolean read FAllowOverflows write FAllowOverflows;
    property    CountOverflows : Boolean read FCountOverflows write FCountOverflows  default True;
  end;


  TSignalOctet = record
    Ch1 : TSignal;
    Ch2 : TSignal;
    Ch3 : TSignal;
    Ch4 : TSignal;
    Ch5 : TSignal;
    Ch6 : TSignal;
    Ch7 : TSignal;
    Ch8 : TSignal;
  end;
  TSignalOctetArray = array of TSignalOctet;


  TSignalOctetFifo = class
  private
    FSignals        : TSignalOctetArray; // The data buffer
    FInp            : Integer;           // Current buffer insertion  point
    FOutp           : Integer;           // Current buffer extraction point
    FFreeCount      : Integer;           // Number of free data slots left
    FFillCount      : Integer;           // Number of filled data slots available
    FOverflows      : Integer;           // Number of times Append  was called while the buffer had no free space
    FUnderflows     : Integer;           // Number of tiems GetData was called while the buffer had no items in it
    FLastValue      : TSignalOctet;      // To repeatedly return the last pair when no samples are available
    FCountOverflows : Boolean;           // Set to true to keep count of overflows
  private
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
    function    GetSize      : Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    procedure   Append( const aCh1, aCh2, aCh3, aCh4, aCh5, aCh6, aCh7, aCh8: TSignal);                        overload;
    procedure   Append( const anOctet: TSignalOctet);                                                          overload;
    function    GetData: TSignalOctet;
    procedure   ClearOverUnderFlows;
  public
    property    Overflows     : Integer read GetOverflows;
    property    Underflows    : Integer read GetUnderflows;
    property    FillCount     : Integer read FFillCount;
    property    FreeCount     : Integer read FFreeCount;
    property    Size          : Integer read GetSize;
    property    CountOverflows: Boolean read FCountOverflows write FCountOverflows  default True;
  end;


  TBytesArray = array of TBytes;
  TBytesFifo  = class
  private
    FBuffer         : TBytesArray;
    FInPointer      : Integer;
    FOutPointer     : Integer;
    FFreeCount      : Integer;
    FFillCount      : Integer;
    FOverflows      : Integer;
    FUnderflows     : Integer;
    FCountOverflows : Boolean;
  private
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
    function    GetSize      : Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    procedure   Append( const aValue: TBytes);
    function    GetData: TBytes;
    procedure   ClearOverUnderFlows;
  public
    property    Overflows     : Integer read GetOverflows;
    property    Underflows    : Integer read GetUnderflows;
    property    FillCount     : Integer read FFillCount;
    property    FreeCount     : Integer read FFreeCount;
    property    Size          : Integer read GetSize;
    property    CountOverflows: Boolean read FCountOverflows write FCountOverflows  default True;
  end;


  TMedianCalculator = class
  private
    FSignals    : TSignalArray;  // The data buffer
    FSorted     : TSignalArray;  // The sorted buffer
    FInp        : Integer;       // Current buffer insertion  point
    FOutp       : Integer;       // Current buffer extraction point
    FIsSorted   : Boolean;
    FSum        : TSignal;       // Cached sum value
  private
    procedure   CheckSorted;
    function    GetCount   : Integer;
    function    GetLowest  : TSignal;
    function    GetMedian  : TSignal;
    function    GetHighest : TSignal;
    function    GetAverage : TSignal;
  private
    procedure   QuickSort;                                                                                     overload;
    procedure   QuickSort( const aValues: TSignalArray; L, R: Integer);                                        overload;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Append( const aValue: TSignal);
    function    GetData: TSignal;
  public
    property    Count   : Integer read GetCount;
    property    Lowest  : TSignal read GetLowest;
    property    Median  : TSignal read GetMedian;
    property    Highest : TSignal read GetHighest;
    property    Average : TSignal read GetAverage;
  end;


  TRMSCalculator = class
  private
    FAvgTime       : TSignal;
    FIsFast        : Boolean;
    FMean          : TSignal;
    FAvgCoef       : TSignal;
  private
    procedure   Recalculate;
    procedure   SetAvgTime( aValue: TSignal);
  public
    constructor Create;
    procedure   Reset;
    procedure   Tick( anInValue: TSignal             ; var anOutValue: TSignal);                               overload;
    procedure   Tick( anInValueL, anInValueR: TSignal; var anOutValue: TSignal);                               overload;
    procedure   SampleRateChanged;
  public
    property    AvgTime : TSignal read FAvgTime write SetAvgTime;
    property    IsFast  : Boolean read FIsFast  write FIsFast;
  end;


  TDbMinMax = class
  private
    FCallRate   : TSignal;        // Tick calls / sample
    FLowValue   : TSignal;
    FHighValue  : TSignal;
    FValeyDecay : TSignal;
    FPeakDecay  : TSignal;
    FValue      : TSignal;
    FPeak       : TSignal;
    FValey      : TSignal;
    FAbsMin     : TSignal;
    FAbsMax     : TSignal;
  public
    constructor Create( const aLowValue, aHighValue, aValeyDecay, aPeakDecay: TSignal);
    procedure   Tick( const aLinearValue: TSignal);
    procedure   Reset;
    procedure   Clear;
    function    ValueRanged( aMin, aMax: Integer): Integer;
    function    PeakRanged ( aMin, aMax: Integer): Integer;
    function    ValeyRanged( aMin, aMax: Integer): Integer;
  public
    property    CallRate   : TSignal read FCallRate write FCallRate;
    property    LowValue   : TSignal read FLowValue  ;
    property    HighValue  : TSignal read FHighValue ;
    property    ValeyDecay : TSignal read FValeyDecay;
    property    PeakDecay  : TSignal read FPeakDecay ;
    property    Value      : TSignal read FValue     ;
    property    Peak       : TSignal read FPeak      ;
    property    Valey      : TSignal read FValey     ;
    property    AbsMin     : TSignal read FAbsMin    ;
    property    AbsMax     : TSignal read FAbsMax    ;
  end;


  TKnobsNullConverter = class
  private
    FSearched     : string;
    FDefaultUnits : string;
  protected
    FName : string;
    FRate : TSignalRate;
  private
    function    ApplyUnits   ( const aUnits      : string; var aValue: TSignal): Boolean;                       virtual;
    function    StringToValue( const aStringValue: string; var aValue: TSignal): Boolean;                       virtual;
    function    PostProcess  ( const aValue: TSignal): TSignal;                                                 virtual;
  public
    constructor Create;                                                                                         virtual;
    function    PosToHint     ( aPos, aSteps: Integer): string;                                                 virtual;
    function    PosToDisplay  ( aPos, aSteps: Integer): string;                                                 virtual;
    function    PosToValue    ( aPos, aSteps: Integer): TSignal;                                                virtual;
    function    DisplayToValue( const aDisplay: string; var aValue: TSignal): Boolean;
  public
    property    Name         : string read FName;
    property    DefaultUnits : string read FDefaultUnits;
    property    Searched     : string read FSearched     write FSearched;
  end;

  TKnobsConverterClass = class of TKnobsNullConverter;


  TKnobsConverters = class( TStringList)
  private
    function    Converter( const anIndex: string): TKnobsNullConverter;
  public
    constructor Create;
    procedure   RegisterConverter( aClass: TKnobsConverterClass);
    procedure   RegisterConverters( const aClasses: array of TKnobsConverterClass);
    function    PosToHint     ( const aName: string; aPos, aSteps: Integer): string;
    function    PosToDisplay  ( const aName: string; aPos, aSteps: Integer): string;
    function    PosToValue    ( const aName: string; aPos, aSteps: Integer): TSignal;
    function    DisplayToValue( const aName, aDisplay: string; var aValue: TSignal): Boolean;
  end;


  function  ParseTextValue( const aType, aStringValue: string; var aSignal: TSignal): Boolean;

  function  ExpAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  SAttack  ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  ExpDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  SDecay   ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;

  function  RangeMap   ( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal;                inline;
  function  AlmostEqual( const aValue1, aValue2: TSignal; const aDelta: TSignal = DELTA_LIMIT): Boolean;         inline;
  function  Normalize  ( const aValue : TSignal): TSignal;                                                       inline;
  function  NormalizeM ( const aValue : TSignal): TSignal;                                                       inline;
  procedure PNormalize ( var aValue : TSignal);                                                                  inline;
  function  AlmostZero ( const aValue : TSignal; const aDelta: TSignal = DELTA_LIMIT): Boolean;                  inline;
  function  CubeInterp ( const fr, inm1, inp, inp1, inp2: TSignal): TSignal;                                     inline;

  function  SignalTypeToColor( aValue: TSignalType): TColor;

  function  LogicToSignal             ( const aValue : Boolean): TSignal;                                        inline;
  function  SignalToLogic             ( const aValue : TSignal): Boolean;                                        inline;
  function  BinaryToSignal            ( const aValue : TBinary; aBitCount: TBitCount): TSignal;                  inline;
  function  SignalToBinary            ( const aValue : TSignal; aBitCount: TBitCount): TBinary;                  inline;
  function  SignalToMute              ( const aValue : TSignal): TSignal;                                        inline;

  function  UnitsTodB                 ( const aValue : TSignal): TSignal;                                        inline;
  function  dBToUnits                 ( const aValue : TSignal): TSignal;                                        inline;
  function  CentsToMultiplier         ( const aValue : TSignal): TSignal;                                        inline;

  function  NoteNumberToUnits         ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToNoteNumber         ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToFrequency          ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToFrequencyLin       ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToFrequencyCubed     ( const aValue : TSignal): TSignal;                                        inline;
  function  FrequencyToUnits          ( const aValue : TSignal): TSignal;                                        inline;
  function  FrequencyToPhaseDelta     ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToPhaseDelta         ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToPhaseDeltaLin      ( const aValue : TSignal): TSignal;                                        inline;
  function  SampleCountToTime         ( const aValue : TSignal): TSignal;                                        inline;
  function  TimeToSampleCount         ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoFastFrequency   ( const aValue : TSignal): TSignal;                                        inline;
  function  LfoFastFrequencyToUnits   ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoMediumFrequency ( const aValue : TSignal): TSignal;                                        inline;
  function  LfoMediumFrequencyToUnits ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoSlowFrequency   ( const aValue : TSignal): TSignal;                                        inline;
  function  LfoSlowFrequencyToUnits   ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoBPMFrequency    ( const aValue : TSignal): TSignal;                                        inline;
  function  LfoBPMFrequencyToUnits    ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoLinFrequency    ( const aValue : TSignal): TSignal;                                        inline;
  function  LfoLinFrequencyToUnits    ( const aValue : TSignal): TSignal;                                        inline;
  function  FrequencyToLfoPhaseDelta  ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoFastPhaseDelta  ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoMediumPhaseDelta( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoSlowPhaseDelta  ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoBPMPhaseDelta   ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToLfoLinPhaseDelta   ( const aValue : TSignal): TSignal;                                        inline;
  function  SlowTimeToSampleCount     ( const aValue : TSignal): TSignal;                                        inline;
  function  UnitsToSvfFrequency       ( const aValue : TSignal): TSignal;                                        inline;
  function  SignalToStr               ( const aValue: TSignal): string;
  function  TimeToStr                 ( const aValue: TSignal): string;
  function  FreqToStr                 ( const aValue: TSignal): string;
  function  FreqToBPMStr              ( const aValue: TSignal): string;
  function  RateMultiplierToStr       ( const aValue: TSignal): string;
  function  RateMultiplierValue       ( const aValue: TSignal): TSignal;

  procedure RegisterConverters( const aClasses: array of TKnobsConverterClass);
  function  Converters: TKnobsConverters;


  function  LookupFrequency          ( const aValue: TSignal): TSignal;
  function  LookupFrequencyLin       ( const aValue: TSignal): TSignal;
  function  LookupFrequencyCubed     ( const aValue: TSignal): TSignal;
  function  LookupPhaseDelta         ( const aValue: TSignal): TSignal;
  function  LookupPhaseDeltaLin      ( const aValue: TSignal): TSignal;
  function  LookupSvfFrequency       ( const aValue: TSignal): TSignal;
  function  LookupLfoSlowFrequency   ( const aValue: TSignal): TSignal;
  function  LookupLfoSlowPhaseDelta  ( const aValue: TSignal): TSignal;
  function  LookupLfoMediumFrequency ( const aValue: TSignal): TSignal;
  function  LookupLfoMediumPhaseDelta( const aValue: TSignal): TSignal;
  function  LookupLfoFastFrequency   ( const aValue: TSignal): TSignal;
  function  LookupLfoFastPhaseDelta  ( const aValue: TSignal): TSignal;
  function  LookupLfoBPMFrequency    ( const aValue: TSignal): TSignal;
  function  LookupLfoBPMPhaseDelta   ( const aValue: TSignal): TSignal;
  function  LookupLfoLinFrequency    ( const aValue: TSignal): TSignal;
  function  LookupLfoLinPhaseDelta   ( const aValue: TSignal): TSignal;
  function  LookupSine               ( const aValue: TSignal): TSignal;
  function  LookupCosine             ( const aValue: TSignal): TSignal;
  function  LookupScaledEnvTime      ( const aValue: TSignal): TSignal;
  function  LookupScaledSEnvTime     ( const aValue: TSignal): TSignal;
  function  LookupEnvTime            ( aRange: TEnvRange; aValue: TSignal): TSignal;
  function  LookupCentsMultiplier    ( const aValue: TSignal): TSignal;

  function  MidiFilterToStr( const aValue: TMidiFilter): string;

  procedure ComputeTables;
  procedure SetSystemRate( aValue: Integer);

  function  SignalToFractional( aValue: TSignal): Integer; inline;
  function  FractionalToSignal( aValue: Integer): TSignal; inline;

  function  SignalPairArrayToStr( const aValue: TSignalPairArray): string;
  function  StrToSignalPairArray( const aValue: string; DoFixup: Boolean; aLowX, aHiX: TSignal): TSignalPairArray;
  function  InterpolateSignalPairArray( const aData: TSignalPairArray; anX: TSignal): TSignal;


var

  ColorControl    : TColor = clBlue;
  ColorAudio      : TColor = clRed;
  ColorLogic      : TColor = clYellow;
  ColorContrLogic : TColor = clGreen;
  ColorUnknown    : TColor = clAqua;
  ColorHighlight  : TColor = clWhite;


implementation


var

  LocConverters : TKnobsConverters;



{ ========
  TSignals = class
  public
    FSignals : TSignalArray;
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal; default;
  private
}

    function    TSignalArrayWrapper.GetCount: Integer;
    begin
      Result := Length( FSignals);
    end;


    function    TSignalArrayWrapper.GetSignal( anIndex: Integer): TSignal;
    begin
      Result := FSignals[ anIndex];
    end;


    procedure   TSignalArrayWrapper.SetSignal( anIndex: Integer; const aValue: TSignal);
    begin
      FSignals[ anIndex] := aValue;
    end;


//  public

    constructor TSignalArrayWrapper.Create( aCount: Integer); // overload;
    begin
      inherited Create;
      SetLength( FSignals, aCount);
    end;


    destructor  TSignalArrayWrapper.Destroy; // override;
    begin
      Clear;
      inherited;
    end;


    procedure   TSignalArrayWrapper.Clear;
    begin
      SetLength( FSignals, 0);
    end;


    procedure   TSignalArrayWrapper.Erase;
    var
      i : Integer;
    begin
      for i := 0 to Length( FSignals) - 1
      do FSignals[ i] := 0.0;
    end;


    procedure   TSignalArrayWrapper.AddSignal( const aValue: TSignal);
    begin
      SetLength( FSignals, Count + 1);
      FSignals[ Count - 1] := aValue;
    end;


    procedure   TSignalArrayWrapper.SetSignals( aStartIndex: Integer; const aValue: TSignalArrayWrapper);
    var
      i : Integer;
    begin
      if Assigned( aValue)
      then begin
        SetLength( FSignals, aStartIndex + aValue.Count);

        for i := aStartIndex to Count - 1
        do Signal[ i] := aValue.Signal[ i - aStartIndex];
      end;
    end;


    function    TSignalArrayWrapper.InterpolateL( const anIndex: TSignal): TSignal;
    var
      p : Integer;
      q : Integer;
      d  : TSignal;
    begin
      p := Trunc( anIndex);
      d := anIndex - p;
      q := ( p + 1);

      if q > Count - 1
      then q := 0;

      Result := FSignals[ p] + d * ( FSignals[ q] - FSignals[ p]);
    end;


    function    TSignalArrayWrapper.InterpolateA( const anIndex: TSignal): TSignal;
    var
      p : Integer;
      q : Integer;
      d  : TSignal;
      a  : TSignal;
    begin
      p := Trunc( anIndex);
      d := anIndex - p;

      if d < 0.5                 // best 0.5 <= d < 1.5
      then begin
        Inc( p);
        d := d + 1;

        if p > Count - 1
        then p := p - Count;
      end;

      a := ( 1 - d) / ( 1 + d);
      q := ( p + 1);

      if q > Count - 1
      then q := q - Count;

      FlastOutput := Normalize( FSignals[ p] + a * ( FSignals[ q] - FLastOutput));
      Result      := FlastOutput;
    end;



{ ========
  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

  TSignalPairs = array of TSignalPair;

  TSignalpairFifo = class
  private
    FSignals        : TSignalPairArray; // The data buffer
    FInp            : Integer;          // Current buffer insertion  point
    FOutp           : Integer;          // Current buffer extraction point
    FFreeCount      : Integer;          // Number of free data slots left
    FFillCount      : Integer;          // Number of filled data slots available
    FAllowOverflows : Boolean;          // True when overflows are allowed, pld data is sicarded then
    FOverflows      : Integer;          // Number of times Append  was called while the buffer had no free space
    FUnderflows     : Integer;          // Number of tiems GetData was called while the buffer had no items in it
    FLastValue      : TSignalPair;      // To repeatedly return the last pair when no samples are available
    FCountOverflows : Boolean;          // Set to true to keep count of overflows
  public
    property    Overflows      : Integer read GetOverflows;
    property    Underflows     : Integer read GetUnderflows;
    property    FillCount      : Integer read FFillCount;
    property    FreeCount      : Integer read FFreeCount;
    property    Size           : Integer read GetSize;
    property    AllowOverflows : Boolean read FAllowOverflows write FAllowOverflows;
    property    CountOverflows : Boolean read FCountOverflows write FCountOverflows  default True;
  private
}


    function    TSignalPairFifo.GetOverflows: Integer;
    begin
      Result     := FOverflows;
      FOverflows := FOverflows - Result;
    end;


    function    TSignalPairFifo.GetUnderflows: Integer;
    begin
      Result      := FUnderflows;
      FUnderflows := FUnderflows - Result;
    end;


    function    TSignalPairFifo.GetSize: Integer;
    begin
      Result := Length( FSignals);
    end;


//  public

    constructor TSignalPairFifo.Create( aSize: Integer);
    begin
      inherited Create;
      CountOverflows := True;
      SetLength( FSignals, aSize);
      Clear;
    end;


    destructor  TSignalPairFifo.Destroy; // override;
    begin
      Clear;
      FFreeCount := 0;
      inherited;
    end;


    procedure   TSignalPairFifo.Clear;
    begin
      FFillCount       := 0;
      FFreeCount       := 0;
      FInp             := 0;
      FOutp            := 0;
      FUnderflows      := 0;
      FOverflows       := 0;
      FLastValue.Left  := 0;
      FLastValue.Right := 0;
      FFreeCount       := Size;
    end;


    procedure   TSignalPairFifo.Append( const aLeft, aRight: TSignal); // overload;
    var
      aPair: TSignalPair;
    begin
      aPair.Left  := aLeft;
      aPair.Right := aRight;
      Append( aPair);
    end;


    procedure   TSignalPairFifo.Append( const aPair: TSignalPair); // overload;
    // When the queue is full discard sample pair, and incremeent the oveflow counteer
    begin
      if FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FSignals[ FInp] := aPair;
        FInp := ( FInp + 1) mod Size;
        Inc( FFillCount);
      end
      else if AllowOverflows
      then begin
        FSignals[ FInp] := aPair;
        FInp  := ( FInp  + 1) mod Size;
        FOutp := ( FOutp + 1) mod Size;
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FOverflows) {$Q+R+};
      end;
    end;


    function    TSignalPairFifo.GetData: TSignalPair;
    // When no sample pairs are available repeatedly return the last pair, and increment the underflow counter.
    begin
      if FFillCount > 0
      then begin
        Dec( FFillCount);
        FLastValue := FSignals[ FOutp];
        FOutp      := ( FOutp + 1) mod Size;
        Inc( FFreeCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FUnderflows) {$Q+R+};
      end;

      Result := FLastValue;
    end;


    procedure   TSignalPairFifo.ClearOverUnderFlows;
    begin
      FUnderflows := 0;
      FOverflows  := 0;
    end;


{
  TSignalOctetFifo = class
  private
    FSignals        : TSignalOctetArray; // The data buffer
    FInp            : Integer;           // Current buffer insertion  point
    FOutp           : Integer;           // Current buffer extraction point
    FFreeCount      : Integer;           // Number of free data slots left
    FFillCount      : Integer;           // Number of filled data slots available
    FOverflows      : Integer;           // Number of times Append  was called while the buffer had no free space
    FUnderflows     : Integer;           // Number of tiems GetData was called while the buffer had no items in it
    FLastValue      : TSignalOctet;      // To repeatedly return the last octet when no samples are available
    FCountOverflows : Boolean;           // Set to true to keep count of overflows
  public
    property    Overflows     : Integer read GetOverflows;
    property    Underflows    : Integer read GetUnderflows;
    property    FillCount     : Integer read FFillCount;
    property    FreeCount     : Integer read FFreeCount;
    property    Size          : Integer read GetSize;
    property    CountOverflows: Boolean read FCountOverflows write FCountOverflows  default True;
  private
}

    function    TSignalOctetFifo.GetOverflows: Integer;
    begin
      Result     := FOverflows;
      FOverflows := FOverflows - Result;
    end;


    function    TSignalOctetFifo.GetUnderflows: Integer;
    begin
      Result      := FUnderflows;
      FUnderflows := FUnderflows - Result;
    end;


    function    TSignalOctetFifo.GetSize: Integer;
    begin
      Result := Length( FSignals);
    end;


//  public

    constructor TSignalOctetFifo.Create( aSize: Integer);
    begin
      inherited Create;
      CountOverflows := True;
      SetLength( FSignals, aSize);
      Clear;
    end;


    destructor  TSignalOctetFifo.Destroy; // override;
    begin
      Clear;
      FFreeCount := 0;
      inherited;
    end;


    procedure   TSignalOctetFifo.Clear;
    begin
      FFillCount       := 0;
      FFreeCount       := 0;
      FInp             := 0;
      FOutp            := 0;
      FUnderflows      := 0;
      FOverflows       := 0;
      FLastValue.Ch1   := 0;
      FLastValue.Ch2   := 0;
      FLastValue.Ch3   := 0;
      FLastValue.Ch4   := 0;
      FLastValue.Ch5   := 0;
      FLastValue.Ch6   := 0;
      FLastValue.Ch7   := 0;
      FLastValue.Ch8   := 0;
      FFreeCount       := Size;
    end;


    procedure   TSignalOctetFifo.Append( const aCh1, aCh2, aCh3, aCh4, aCh5, aCh6, aCh7, aCh8: TSignal); // overload;
    var
      anOctet: TSignalOctet;
    begin
      anOctet.Ch1 := aCh1;
      anOctet.Ch2 := aCh2;
      anOctet.Ch3 := aCh3;
      anOctet.Ch4 := aCh4;
      anOctet.Ch5 := aCh5;
      anOctet.Ch6 := aCh6;
      anOctet.Ch7 := aCh7;
      anOctet.Ch8 := aCh8;
      Append( anOctet);
    end;


    procedure   TSignalOctetFifo.Append( const anOctet: TSignalOctet); // overload;
    // When the queue is full discard sample octet, and increment the oveflow counteer
    begin
      if FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FSignals[ FInp] := anOctet;
        FInp := ( FInp + 1) mod Size;
        Inc( FFillCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FOverflows) {$Q+R+};
      end;
    end;


    function    TSignalOctetFifo.GetData: TSignalOctet;
    // When the queue is empty repeatedly return the last octet, and increment the underflow counter.
    begin
      if FFillCount > 0
      then begin
        Dec( FFillCount);
        FLastValue := FSignals[ FOutp];
        FOutp      := ( FOutp + 1) mod Size;
        Inc( FFreeCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FUnderflows) {$Q+R+};
      end;

      Result := FLastValue;
    end;


    procedure   TSignalOctetFifo.ClearOverUnderFlows;
    begin
      FUnderflows := 0;
      FOverflows  := 0;
    end;


{ ========
  TBytesArray = array of Bytes;
  TBytesFifo = class
  private
    FBuffer         : TBytesArray;
    FInPointer      : Integer;
    FOutPounter     : Integer;
    FFreeCount      : Integer;
    FFillCount      : Integer;
    FOverflows      : Integer;
    FUnderflows     : Integer;
    FCountOverflows : Boolean;
  public
    property    Overflows     : Integer read GetOverflows;
    property    Underflows    : Integer read GetUnderflows;
    property    FillCount     : Integer read FFillCount;
    property    FreeCount     : Integer read FFreeCount;
    property    Size          : Integer read GetSize;
    property    CountOverflows: Boolean read FCountOverflows write FCountOverflows  default True;
  private
}

    function    TBytesFifo.GetOverflows: Integer;
    begin
      Result     := FOverflows;
      FOverflows := FOverflows - Result;
    end;


    function    TBytesFifo.GetUnderflows: Integer;
    begin
      Result      := FUnderflows;
      FUnderflows := FUnderflows - Result;
    end;


    function    TBytesFifo.GetSize: Integer;
    begin
      Result := Length( FBuffer);
    end;


//  public

    constructor TBytesFifo.Create( aSize: Integer);
    begin
      inherited Create;
      CountOverflows := True;
      SetLength( FBuffer, aSize);
      Clear;
    end;


    destructor  TBytesFifo.Destroy; // override;
    begin
      Clear;
      FFreeCount := 0;
      inherited;
    end;


    procedure   TBytesFifo.Clear;
    begin
      FFreeCount  := 0;
      FFillCount  := 0;
      FFreeCount  := 0;
      FInPointer  := 0;
      FOutPointer := 0;
      FUnderflows := 0;
      FOverflows  := 0;
      FFreeCount  := Size;
    end;


    procedure   TBytesFifo.Append( const aValue: TBytes);
    begin
      if FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FBuffer[ FInPointer] := aValue;
        FInPointer           := ( FInPointer + 1) mod Size;
        Inc( FFillCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FOverflows) {$Q+R+};
      end;
    end;


    function    TBytesFifo.GetData: TBytes;
    begin
      if FFillCount > 0
      then begin
        Dec( FFillCount);
        Result      := FBuffer[ FOutPointer];
        FOutPointer := ( FOutPointer + 1) mod Size;
        Inc( FFreeCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FUnderflows) {$Q+R+};
      end;
    end;


    procedure   TBytesFifo.ClearOverUnderFlows;
    begin
      FOverflows  := 0;
      FUnderflows := 0;
    end;


{ ========
  TMedianCalculator = class
  private
    FSignals    : TSignalArray;  // The data buffer
    FSorted     : TSignalArray;  // The sorted buffer
    FInp        : Integer;       // Current buffer insertion  point
    FOutp       : Integer;       // Current buffer extraction point
    FIsSorted   : Boolean;
    FSum        : TSignal;       // Cached sum value
  public
    property    Count   : Integer read GetCount;
    property    Lowest  : TSignal read GetLowest;
    property    Median  : TSignal read GetMedian;
    property    Highest : TSignal read GetHighest;
    property    Average : TSignal read GetAverage;
  private
}

    procedure   TMedianCalculator.CheckSorted;
    var
      i : Integer;
    begin
      if ( Count > 0) and not FIsSorted
      then begin
        for i := 0 to Count - 1
        do FSorted[ i] := FSignals[ i];

        QuickSort;
        FIsSorted := True;
      end;
    end;


    function    TMedianCalculator.GetCount: Integer;
    begin
      if Assigned( FSignals)
      then Result := Length( FSignals)
      else Result := 0;
    end;


    function    TMedianCalculator.GetLowest: TSignal;
    begin
      if Count > 0
      then begin
        CheckSorted;
        Result := FSorted[ 0];
      end
      else Result := 0;
    end;


    function    TMedianCalculator.GetMedian: TSignal;
    begin
      if Count > 0
      then begin
        CheckSorted;

        if Odd( Count)
        then Result := FSorted[ Count shr 1]
        else Result := ( FSorted[( Count shr 1) - 1] + FSorted[ Count shr 1]) * 0.5;
      end
      else Result := 0;
    end;


    function    TMedianCalculator.GetHighest: TSignal;
    begin
      if Count > 0
      then begin
        CheckSorted;
        Result := FSorted[ Count - 1];
      end
      else Result := 0;
    end;


    function    TMedianCalculator.GetAverage : TSignal;
    begin
      if Count > 0
      then Result := FSum / Count
      else Result := 0;
    end;


//  private

    procedure   TMedianCalculator.QuickSort; // overlaod;
    begin
      QuickSort( FSorted, 0, Count - 1);
    end;


    procedure   TMedianCalculator.QuickSort( const aValues: TSignalArray; L, R: Integer); // overlaod;
    var
      I, J, P: Integer;
      aTmp : TSignal;
    begin
      repeat
        I := L;
        J := R;
        P := ( L + R) shr 1;

        repeat
          while aValues[ I] < aValues[ P] do Inc( I);
          while aValues[ J] > aValues[ P] do Dec( J);
          if I <= J
          then begin
            if I <> J
            then begin
              aTmp        := aValues[ I];
              aValues[ I] := aValues[ J];
              aValues[ J] := aTmp;
            end;

            if P = I
            then P := J
            else if P = J
            then P := I;

            Inc( I);
            Dec( J);
          end;
        until I > J;

        if L < J then QuickSort( aValues, L, J);
        L := I;
      until I >= R;
    end;


//  public

    constructor TMedianCalculator.Create( aSize: Integer);
    begin
      inherited Create;
      SetLength( FSignals, aSize);
      SetLength( FSorted , aSize);
    end;


    destructor  TMedianCalculator.Destroy; // override;
    begin
      inherited;
    end;


    procedure   TMedianCalculator.Append( const aValue: TSignal);
    begin
      FSignals[ FInp] := aValue;
      FInp            := ( FInp + 1) mod Count;
      FSum            := FSum + aValue - FSignals[ FInp];
      FIsSorted       := False;
    end;


    function    TMedianCalculator.GetData: TSignal;
    begin
      Result := FSignals[ FOutp];
      FOutp  := ( FOutp + 1) mod Count;
    end;



{ ========
  TRMSCalculator = class
  private
    FAvgTime       : TSignal;
    FIsFast        : Boolean;
    FMean          : TSignal;
    FAvgCoef       : TSignal;
  public
    property    AvgTime : TSignal read FAvgTime write SetAvgTime;
    property    IsFast  : Boolean read FIsFast  write FIsFast;
  orivate
}

    procedure   TRMSCalculator.Recalculate;
    begin;
      if FIsFast
      then FAvgCoef := 1.0 - Exp( - 1.0 / TimeToSampleCount    ( FAvgTime))
      else FAvgCoef := 1.0 - Exp( - 1.0 / SlowTimeToSampleCount( FAvgTime));
    end;


    procedure   TRMSCalculator.SetAvgTime( aValue: TSignal);
    begin
      FAvgTime := aValue;
      Recalculate;
    end;


//  public

    constructor TRMSCalculator.Create;
    begin
      inherited;
      IsFast := False;
      Reset;
      AvgTime := 0.1;              // 100 ms
    end;


    procedure   TRMSCalculator.Reset;
    begin
      FMean := 0.0;
    end;


    procedure   TRMSCalculator.Tick( anInValue: TSignal; var anOutValue: TSignal); // overload;
    begin
      FMean      := FMean + FAvgCoef * ( anInValue * anInValue - FMean);
      anOutValue := Sqrt( FMean);
    end;


    procedure   TRMSCalculator.Tick( anInValueL, anInValueR: TSignal; var anOutValue: TSignal); // overload;
    begin
      FMean      := FMean + FAvgCoef * ( anInValueL * anInValueL + anInValueR * anInValueR - FMean);
      anOutValue := Sqrt( FMean);
    end;


    procedure   TRMSCalculator.SampleRateChanged;
    // Notification for sample rate change, the actual values are in global variables
    // System_Rate for the audio rate and Control_Rate for the control rate
    // Modules caching sample rate dependent data should recalculate such data
    // in an overriden version of SampleRateChanged.
    begin
      Recalculate;
    end;


{ ========
  TDbMinMaxCalculator = class
  private
    FCallRate   : TSignal;        // Tick calls / sample
    FLowValue   : TSignal;
    FHighValue  : TSignal;
    FValeyDecay : TSignal;
    FPeakDecay  : TSignal;
    FValue      : TSignal;
    FPeak       : TSignal;
    FValey      : TSignal;
    FAbsMin     : TSignal;
    FAbsMax     : TSignal;
  public
    property    CallRate   : TSignal read FCallRate write FCallRate;
    property    LowValue   : TSignal read FLowValue  ;
    property    HighValue  : TSignal read FHighValue ;
    property    ValeyDecay : TSignal read FValeyDecay;
    property    PeakDecay  : TSignal read FPeakDecay ;
    property    Value      : TSignal read FValue     ;
    property    Peak       : TSignal read FPeak      ;
    property    Valey      : TSignal read FValey     ;
    property    AbsMin     : TSignal read FAbsMin    ;
    property    AbsMax     : TSignal read FAbsMax    ;
  public
}

    constructor TDbMinMax.Create( const aLowValue, aHighValue, aValeyDecay, aPeakDecay: TSignal);
    const
      OFT = 1.0 / 512.0;
    begin
      inherited Create;
      FCallRate   := OFT;
      FLowValue   := aLowValue;
      FHighValue  := aHighValue;
      FValeyDecay := aValeyDecay;
      FPeakDecay  := aPeakDecay;
      Clear;
    end;


    procedure   TDbMinMax.Tick( const aLinearValue: TSignal);
    begin
      FValue := Normalize( UnitsTodB( aLinearValue));

      if FValue > FPeak
      then FPeak := FValue
      else FPeak := ( FPeakDecay / FCallRate) * ( FValue - FPeak) + FPeak;

      if FValue < FValey
      then FValey := FValue
      else FValey := ( FValeyDecay / FCallRate) * ( FValue - FValey) + FValey;

      if FValue > FAbsMax
      then FAbsMax := FValue;

      if FValue < FAbsMin
      then FAbsMin := FValue;
    end;


    procedure   TDbMinMax.Reset;
    begin
      FPeak   := FValue;
      FValey  := FValue;
      FAbsMin := FValue;
      FAbsMax := FValue;
    end;


    procedure   TDbMinMax.Clear;
    begin
      FAbsMin := FHighValue;
      FAbsMax := FLowValue;
      FValue  := FLowValue;
      FPeak   := FValue;
      FValey  := FValue;
    end;


    function    TDbMinMax.ValueRanged( aMin, aMax: Integer): Integer;
    begin
      Result := Round( RangeMap( Clip( FValue, FLowValue, FHighValue), FLowValue, FHighValue, aMin, aMax));
    end;


    function    TDbMinMax.PeakRanged( aMin, aMax: Integer): Integer;
    begin
      Result := Ceil( RangeMap( Clip( FPeak, FLowValue, FHighValue), FLowValue, FHighValue, aMin, aMax));
    end;


    function    TDbMinMax.ValeyRanged( aMin, aMax: Integer): Integer;
    begin
      Result := Floor( RangeMap( Clip( FValey, FLowValue, FHighValue), FLowValue, FHighValue, aMin, aMax));
    end;



  function  IsTimeValue( const aValue: string): Boolean;
  begin
    Result := Pos( ':', aValue) > 0;
  end;


  function  ParseTextValue( const aType, aStringValue: string; var aSignal: TSignal): Boolean;
  begin
    if Converters <> nil
    then Result := Converters.DisplayToValue( aType, aStringValue, aSignal)
    else begin
      aSignal := UNDEFINED;
      Result  := False;
    end;
  end;


  function  ExpAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ x - 1  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := LookupScaledEnvTime( aTim);
    Result   := Normalize( aStartValue + Unscaled * ( anEndValue - aStartValue));
  end;


  function  LinAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := aTim;
    Result   := Normalize( aStartValue + Unscaled * ( anEndValue - aStartValue));
  end;


  function  LogAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ ( 1 - x)  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - LookupScaledEnvTime( 1 - aTim);
    Result   := Normalize( aStartValue + Unscaled * ( anEndValue - aStartValue));
  end;


  function  SAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = sqr( sin( x * pi / 2))  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := LookupScaledSEnvTime( aTim);
    Result   := Normalize( aStartValue + Unscaled * ( anEndValue - aStartValue));
  end;


  function  ExpDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ ( 1 - x) - 1  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := LookupScaledEnvTime( 1 - aTim);
    Result   := Normalize( anEndValue + Unscaled * ( aStartValue - anEndValue));
  end;


  function  LinDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 1 - x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - aTim;
    Result   := Normalize( anEndValue + Unscaled * ( aStartValue - anEndValue));
  end;


  function  LogDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - LookupScaledEnvTime( aTim);
    Result   := Normalize( anEndValue + Unscaled * ( aStartValue - anEndValue));
  end;


  function  SDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 1 - sqr( sin( x * pi / 2))  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := LookupScaledSEnvTime( 1 - aTim);
    Result   := Normalize( anEndValue + Unscaled * ( aStartValue - anEndValue));
  end;

  function  Normalize( const aValue: TSignal): TSignal; inline;
  //
  // This actually checks if aValue is zero, a denormal or it's absulote value being < 1e-20
  // [or the value not being a number (NaN)] and it treats all those cases as if they were zero.
  // hmm .. Nan check adds like 5% to computtional load ... commented it out. Maybe later I cn
  // come up with a bit pattern check that finds both Nans and Denormals.
  // This applies to 'procedure Normalize();' as well.
  //
  // ok, so what do we actually have ... IEEE 754 double precission:
  //
  //   +-+-----------+-------------------------------------------+
  //   |1|11         |52                                         |
  //   +-+-----------+-------------------------------------------+
  //   | |$7ff       |$fffffffffffff                             |
  //   +-+-----------+-------------------------------------------+
  //   |s|e          |m                                          |
  //   +-+-----------+-------------------------------------------+
  //
  //   normal double       0 < e < $7ff          v = (-1)^s * 2^(e-1023) * 1.m
  //   "denormal"          e = 0    and m <> 0   v = (-1)^s * 2^-1022    * 0.m
  //   Zero                e = 0    and m  = 0   v = (-1)^s * 2^-1022    * 0.m
  //   Infinity (INF)      e = $7ff and m  = 0   v = (-1)^s * INF
  //   Not a Number (NaN)  e = $7ff and m <> 0   NaN
  //
  // Original code here was :
  //
  //   if (( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) or ( Abs( aValue) < NORMAL_LIMIT)
  //   then Result := 0
  //   else Result := aValue;
  //
  // Then tried :
  //
  //   if (( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) or IsNan( aValue) or ( Abs( aValue) < NORMAL_LIMIT)
  //
  //   Which adds about 5% runtime overhead on a 90% patch.
  //
  // sooo ... but ... the solution may be to just sub one from the exponent ... and then checking it to
  // be >= $7fe something like in:
  //
  //   if ((( PWordArray( @ aValue)^[ 3] and $7ff0) - $10) >= $7fe0) or ( Abs( aValue) < NORMAL_LIMIT)
  //
  // Which then should lead to a conversion of all non-normal floats into zero. This adds less than 1% of
  // extra overhead for the same patch.
  //
  type
    TWordArray = array[ 0 .. 3] of Word;
    PWordArray = ^TWordArray;
  begin
{$Q-R-}
//  if (( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) or ( Abs( aValue) < NORMAL_LIMIT)
    if ((( PWordArray( @ aValue)^[ 3] and $7ff0) - $10) >= $7fe0) or ( Abs( aValue) < NORMAL_LIMIT)
    then Result := 0
    else Result := aValue;
{$Q+R+}
  end;


  function  NormalizeM( const aValue: TSignal): TSignal; inline;
  //
  // Normalization for mute signals
  //
  type
    TWordArray = array[ 0 .. 3] of Word;
    PWordArray = ^TWordArray;
  begin
{$Q-R-}
    if ((( PWordArray( @ aValue)^[ 3] and $7ff0) - $10) >= $7fe0) or ( Abs( aValue) < MUTE_LIMIT)
    then Result := 0
    else Result := aValue;
{$Q+R+}
  end;


  procedure PNormalize( var aValue : TSignal); inline;
  // See remarks for 'function Nirmalize();' above.
  type
    TWordArray = array[ 0 .. 3] of Word;
    PWordArray = ^TWordArray;
  begin
//  if (( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) (* or IsNan( aValue) *) or ( Abs( aValue) < NORMAL_LIMIT)
    if ((( PWordArray( @ aValue)^[ 3] and $7ff0) - $10) >= $7fe0) or ( Abs( aValue) < NORMAL_LIMIT)
    then aValue := 0;
  end;


  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    Result := anOutLow + (( aValue - anInLow) / ( anInHigh - anInLow)) * ( anOutHigh - anOutLow);
  end;


  function  AlmostEqual( const aValue1, aValue2: TSignal; const aDelta: TSignal = DELTA_LIMIT): Boolean;  inline;
  begin
    Result := Abs( aValue1 - aValue2) < aDelta;
  end;


  function  AlmostZero( const aValue: TSignal; const aDelta: TSignal = DELTA_LIMIT): Boolean; inline;
  begin
    Result := Abs( aValue) < aDelta;
  end;


  function  CubeInterp( const fr, inm1, inp, inp1, inp2: TSignal): TSignal; inline;
  begin
    Result :=
      inp + 0.5 * fr * ( inp1 - inm1 + fr * ( 4.0 * inp1 + 2.0 * inm1 - 5.0 * inp - inp2 + fr * ( 3.0 * ( inp - inp1) - inm1 + inp2)));
  end;


  function  SignalTypeToColor( aValue: TSignalType): TColor;
  begin
    case aValue of
      stControl    : Result := ColorControl;
      stAudio      : Result := ColorAudio;
      stLogic      : Result := ColorLogic;
      stContrLogic : Result := ColorContrLogic;
      else           Result := ColorUnknown;
    end;
  end;


  function  LogicToSignal( const aValue: Boolean): TSignal; inline;
  begin
    if aValue
    then Result := 1
    else Result := -1;
  end;


  function  SignalToLogic( const aValue: TSignal): Boolean; inline;
  begin
    Result := aValue > 0;
  end;


  function  BinaryToSignal( const aValue : TBinary; aBitCount: TBitCount): TSignal; inline;
  begin
    Result := ( 1.0 * aValue) / ( 1 shl ( aBitCount - 1));
  end;

  function  SignalToBinary( const aValue : TSignal; aBitCount: TBitCount): TBinary; inline;
  begin
    Result := Round( Clip( aValue, -1, 1) * ( 1 shl ( aBitCount - 1)));
  end;


  function  SignalToMute( const aValue  : TSignal): TSignal; inline;
  type
    TWordArray = array[ 0 .. 3] of Word;
    PWordArray = ^TWordArray;
  begin
    if (( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) or ( Abs( aValue) < MUTE_LIMIT)
    then Result := 1.0
    else Result := Normalize( 1 - aValue);
  end;


  function  UnitsTodB( const aValue : TSignal): TSignal; inline;
  // todo : make a LUT
  begin
    if aValue <= MIN_DB_SIGNAL
    then Result := - MAX_DB_VAL
    else Result := 20 * Log10( aValue);
  end;


  function  dBToUnits( const aValue : TSignal): TSignal; inline;
  // todo : make a LUT
  const
    OT = 1.0 / 20.0;
  begin
    Result := Power( 10, aValue * OT);
  end;


  function  CentsToMultiplier( const aValue : TSignal): TSignal; inline;
  begin
    Result := Power( 2, aValue / 1200.0);
  end;


  function  CentsToMultiplierScaled( const aValue : TSignal): TSignal; inline;
  begin
    Result := CentsToMultiplier( aValue * CENTS);
  end;


  function  UnitsToNoteNumber( const aValue: TSignal): TSignal; inline;
  begin
    Result := NOTE_SCALING * aValue;
  end;


  function  NoteNumberToUnits( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * NOTE_SCALING_REC;
  end;


  function  UnitsToFrequency( const aValue: TSignal): TSignal; inline;
  // such that note Nr 69 maps to A 440 Hz
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // freq = ( 2 ^ ( notenr - 69) / 12) * 440
  // Result := Power( 2, ( aValue - 69) / 12) * ReferenceA;
  begin
    Result := Power( 2, ( UnitsToNoteNumber( aValue) - MiddleNote) * NotesPerOctaveRec) * ReferenceA;
  end;


  function  UnitsToFrequencyLin( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * 10000;
  end;


  function  UnitsToFrequencyCubed( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * aValue * aValue * 10000;
  end;


  function  FrequencyToUnits( const aValue: TSignal): TSignal; inline;
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // n  = 12 * log2( f / 440) + 69
  //  Result := Round( 12 * Log2( aValue / ReferenceA)) + 69;
  begin
    Result := NoteNumberToUnits( NotesPerOctave * Log2( aValue / ReferenceA) + MiddleNote);
  end;


  function  FrequencyToPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * System_Rate_Rec;
  end;


  function  UnitsToPhaseDelta( const aValue : TSignal): TSignal; inline;
  begin
    Result := FrequencyToPhaseDelta( UnitsToFrequency( aValue));
  end;


  function  UnitsToPhaseDeltaLin( const aValue : TSignal): TSignal; inline;
  begin
    Result := FrequencyToPhaseDelta( UnitsToFrequencyLin( aValue));
  end;


  function  SampleCountToTime( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * System_Rate_Rec;
  end;


  function  TimeToSampleCount( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * System_Rate;
  end;


  // ------------------------------------

  function  UnitsToLfoFastFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 43) * ONE_TWELFTH) * 3.05);
  end;


  function  LfoFastFrequencyToUnits( const aValue : TSignal): TSignal; // inline;
  begin
    Result := NoteNumberToUnits( 12 * Log2( aValue / 3.05) + 43);
  end;


  function  UnitsToLfoMediumFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 103) * ONE_TWELFTH) * 3.05);
  end;


  function  LfoMediumFrequencyToUnits( const aValue: TSignal): TSignal; // inline;
  begin
    Result := NoteNumberToUnits( 12 * Log2( aValue / 3.05) + 103);
  end;


  function  UnitsToLfoSlowFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 163) * ONE_TWELFTH) * 3.05);
  end;


  function  LfoSlowFrequencyToUnits( const aValue : TSignal): TSignal; // inline;
  begin
    Result := NoteNumberToUnits( 12 * Log2( aValue / 3.05) + 163);
  end;


  function  UnitsToLfoBPMFrequency( const aValue: TSignal): TSignal; // inline;
  const
    OT = 1.0 / 30.0;
  begin
    Result := Normalize( UnitsToNoteNumber( aValue) * OT);
  end;


  function  LfoBPMFrequencyToUnits( const aValue : TSignal): TSignal; // inline;
  begin
    Result := NoteNumberToUnits( aValue * 30.0);
  end;


  function  UnitsToLfoLinFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := aValue * 15;
  end;


  function  LfoLinFrequencyToUnits( const aValue : TSignal): TSignal; // inline;
  const
    OT = 1.0 / 15.0;
  begin
    Result := aValue * OT;
  end;


  function  FrequencyToLfoPhaseDelta( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( CONTROL_DEC * aValue * System_Rate_Rec);
  end;


  function  UnitsToLfoFastPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoFastFrequency( aValue));
  end;


  function  UnitsToLfoMediumPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoMediumFrequency( aValue));
  end;


  function  UnitsToLfoSlowPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoSlowFrequency( aValue));
  end;


  function  UnitsToLfoBPMPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoBPMFrequency( aValue));
  end;


  function  UnitsToLfoLinPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoLinFrequency( aValue))
  end;


  function  SlowTimeToSampleCount( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * Control_Rate;
  end;


  function  UnitsToSvfFrequency( const aValue: TSignal): TSignal; inline;
  // NOTE: this is based on an SVF sampled at twice the system rate
  begin
    Result := Normalize( 2 * Sin( PI * Min( 0.25, ( UnitsToFrequency( aValue) * System_Rate_2_Rec))));
  end;


  function  UnitsToSine( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( Sin( TWO_PI * aValue));
  end;


  function  UnitsToCosine( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( Cos( TWO_PI * aValue));
  end;


  function  UnitsToScaledEnvTime( const aValue: TSignal): TSignal;
  const
    OTTT = 1.0 / 1023.0;
  begin
    Result := Normalize(( Power( 2, 10 * aValue) - 1 ) * OTTT);
  end;


  function  UnitsToScaledSEnvTime( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( Sqr( Sin( HALF_PI * aValue)));
  end;


  function  TimeToSeconds( const aValue: string; const aDefault: TSignal): TSignal;
  var
    aParts   : TStringList;
    anHours  : Integer;
    aMinutes : Integer;
    aSeconds : Integer;
  begin
    Result    := aDefault;
    anHours   := -1;
    aMinutes  := -1;
    aSeconds  := -1;
    aParts := Explode( aValue, ':');

    try
      if aParts.Count = 2
      then begin
        aSeconds := StrToIntDef( aParts[ 1], -1);
        aMinutes := StrToIntDef( aParts[ 0], -1);
        anHours  := 0;
      end
      else if aParts.Count = 3
      then begin
        aSeconds := StrToIntDef( aParts[ 2], -1);
        aMinutes := StrToIntDef( aParts[ 1], -1);
        anHours  := StrToIntDef( aParts[ 0], -1);
      end;

      if ( aSeconds >= 0) and ( aMinutes >= 0) and ( anHours >= 0)
      then Result := aSeconds + 60 * ( aMinutes + 60 * anHours);
    finally
      aParts.DisposeOf;
    end;
  end;


  function  NoteNameToUnits( const aNoteName: string; const aDefault: TSignal): TSignal;
  // C[#][-]0
  var
    Base  : Integer;
    Sharp : Integer;
    Sign  : Integer;
    N     : Char;
    Shift : Integer;
    p     : Integer;
    T     : string;
    NN    : Integer;
  begin
    Shift :=  0;
    Sharp :=  0;
    Sign  :=  1;
    Base  := -1;

    if Length( aNoteName) > 0
    then begin
      N := UpCase( aNoteName[ 1]);

      case N of
        'C' : Base := 12     ;
        'D' : Base := 12 +  2;
        'E' : Base := 12 +  4;
        'F' : Base := 12 +  5;
        'G' : Base := 12 +  7;
        'A' : Base := 12 +  9;
        'B' : Base := 12 + 11;
        else  Base := -1;
      end;

      if ( Length( aNoteName) > 1) and ( Base > 0)
      then begin
        if aNoteName[ 2] = '#'
        then begin
          Sharp := 1;
          p     := 2;
        end
        else p := 1;
      end
      else p := 0;

      if ( Length( aNoteName) > p) and ( base > 0)
      then begin
        if aNoteName[ p + 1] = '-'
        then begin
          Sign := -1;
          p    := p + 1;
        end;
      end;

      T := Copy( aNoteName, p + 1, Length( aNoteName));

      if T[ Length( T)] = '#'
      then begin
        Sharp := 1;
        T     := Copy( T, 1, Length( T) - 1);
      end;

      Shift := StrToIntDef( T, 0);
    end;

    if Base > 0
    then begin
      NN     := Base + Sharp + Sign * Shift * 12;
      Result := NoteNumberToUnits( NN);
    end
    else Result := aDefault;
  end;


  function  UnitsToNoteName( const aValue: TSignal; Compact: Boolean = True): string;
  const
    Notes  : array[ 0 .. 11] Of Char = 'CCDDEFFGGAAB';
    Sharps : array[ 0 .. 11] Of Char = ' # #  # # # ';
  var
    i          : Integer;
    S          : string;
    aNoteValue : Integer;
  begin
    if IsNan( aValue)
    then aNoteValue := Round( MiddleNote)
    else aNoteValue := Round( UnitsToNoteNumber( aValue));
    i          := MathIntMod( aNoteValue, 12);
    S          := Sharps[ i];

    if Compact and ( S = ' ')
    then S := '';

    Result := Format( '%s%s%d', [ Notes[ i], S, ( MathIntDiv( aNoteValue, 12)) - 1], AppLocale);
  end;


  function  UnitsToNoteName12( const aValue: TSignal): string;
  const
    Names : array[ 0 .. 11] of string = (
      'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'
    );
  var
    aNote : Integer;
  begin
    aNote := Round( 12 * aValue);

    if aNote in [ 0 .. 11]
    then Result := Names[ aNote]
    else Result := 'inv';
  end;


  function  IsNoteName( const aNoteName: string): Boolean;
  // C[#][-]0[#]
  var
    p : Integer;
    i : Integer;
  begin
    Result := False;

    if Length( aNoteName) > 0
    then begin
      Result := CharInSet( aNoteName[ 1], ['a' .. 'g', 'A' .. 'G']);

      if Result and ( Length( aNoteName) > 1)
      then begin
        if aNoteName[ 2] = '#'
        then p := 2
        else p := 1;
      end
      else p := 0;

      if Result and ( Length( aNoteName) > p) and ( aNoteName[ p + 1] = '-')
      then p := p + 1;

      if Result
      then begin
        for i := p + 1 to Length( aNoteName)
        do begin
          if not CharInset( aNoteName[ i], [ '0' .. '9'])
          then begin
            if ( aNoteName[ i] = '#') and ( i = Length( aNoteName))
            then Break
            else begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;
  end;


  function  SignalToStr( const aValue: TSignal): string;
  begin
    Result := FloatToStrF( aValue, ffGeneral, 15, 0, AppLocale);
  end;


  function  TimeToStr( const aValue: TSignal): string;
  const
    DAY     = 24.0 * 3600.0;
    DAY_REC = 1.0 / DAY;
  var
    Days  : TSignal;
    Value : TSignal;
  begin
    if aValue = 0
    then Result := '0 s'
    else if aValue < 0.0001
    then Result := Format( '%.1f s', [ aValue * 1e6], AppLocale)
    else if aValue < 0.001
    then Result := Format( '%.0f s', [ aValue * 1e6], AppLocale)
    else if aValue < 0.01
    then Result := Format( '%.2f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 0.1
    then Result := Format( '%.1f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 1
    then Result := Format( '%.0f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 10
    then Result := Format( '%.2f s' , [ aValue      ], AppLocale)
    else if aValue < 60
    then Result := Format( '%.1f s' , [ aValue      ], AppLocale)
    else if aValue < 3600
    then Result := FormatDateTime( 'nn:ss'  , aValue * DAY_REC, AppLocale)
    else if aValue < DAY
    then Result := FormatDateTime( 'h:nn:ss'   , aValue * DAY_REC, AppLocale)
    else begin
      Value  := FloatMod( aValue, DAY);
      Days   := aValue / DAY;
      Result := Format( '%.0f-%s', [ Days, FormatDateTime( 'hh:nn:ss', Value * DAY_REC, AppLocale)], AppLocale);
    end;
  end;


  function  FreqToStr( const aValue: TSignal): string;
  begin
    if aValue = 0
    then Result := '0 Hz'
    else if Abs( aValue) < 0.1
    then Result := TimeToStr( 1 / aValue)
    else if Abs( aValue) < 1
    then Result := Format( '%.3f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 10
    then Result := Format( '%.3f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 100
    then Result := Format( '%.2f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 1000
    then Result := Format( '%.1f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 10000
    then Result := Format( '%.2f kHz', [ aValue * 1e-3], AppLocale)
    else Result := Format( '%.1f kHz', [ aValue * 1e-3], AppLocale);
  end;


  function  FreqToBPMStr( const aValue: TSignal): string;
  begin
    Result := Format( '%.0f', [ aValue * 60], AppLocale);
  end;


  function  RateMultiplierToStr( const aValue: TSignal): string;
  var
    aVal : Integer;
  begin
    aVal := Round( aValue);
    if aVal < 0
    then Result := Format( '1 / %d', [ 1 - aVal])
    else Result := Format( '%d'    , [ 1 + aVal]);
  end;


  function  RateMultiplierValue( const aValue: TSignal): TSignal;
  var
    aVal : Integer;
  begin
    aVal := Round( aValue);
    if aVal < 0
    then Result := 1.0 / ( 1 - aVal)
    else Result :=         1 + aVal;
  end;


  procedure RegisterConverters( const aClasses: array of TKnobsConverterClass);
  begin
    LocConverters.RegisterConverters( aClasses);
  end;


  function  Converters: TKnobsConverters;
  begin
    Result := LocConverters;
  end;



const

  LutRange = 8192;          // [ 0, 1] TSignal range maps to [0        , LutRange] table range
  LutSize  = 2 * LutRange;  // [-1, 1] TSignal range maps to [-LutRange, LutRange] table range

type

  TLut = array[ - LutRange - 1 .. LutRange + 1] of TSignal;  // -/+ 1 to allow for interpolation without checking bounds on first/last table element

var

  UnitsToFrequencyTable           : TLut;
  UnitsToFrequencyLinTable        : TLut;
  UnitsToFrequencyCubedTable      : TLut;
  UnitsToPhaseDeltaTable          : TLut;
  UnitsToPhaseDeltaLinTable       : TLut;
  UnitsToSvfFreqTable             : TLut;
  UnitsToLfoSlowFrequencyTable    : TLut;
  UnitsToLfoSlowPhaseDeltaTable   : TLut;
  UnitsToLfoMediumFrequencyTable  : TLut;
  UnitsToLfoMediumPhaseDeltaTable : TLut;
  UnitsToLfoFastFrequencyTable    : TLut;
  UnitsToLfoFastPhaseDeltaTable   : TLut;
  UnitsToLfoBPMFrequencyTable     : TLut;
  UnitsToLfoBPMPhaseDeltaTable    : TLut;
  UnitsToLfoLinFrequencyTable     : TLut;
  UnitsToLfoLinPhaseDeltaTable    : TLut;
  UnitsToSineTable                : TLut;
  UnitsToCosineTable              : TLut;
  UnitsToScaledEnvTimeTable       : TLut;
  UnitsToScaledSEnvTimeTable      : TLut;
  CentsToMultiplierTable          : TLut;


  procedure ComputeTable( var aTable: TLut; const aFunc: TSignalFunction);
  var
    i : Integer;
  begin
    for i := Low( aTable) to High( aTable)
    do aTable[ i] := Normalize( aFunc( i / LutRange));
  end;


  function TableLookup( const aValue: TSignal; const aTable : TLut): TSignal; inline;
  var
    aLow   : Integer;
    aVal   : TSignal;
    aDelta : TSignal;
  begin
    aVal   := Clip( aValue * LutRange, Low( aTable), High( aTable) - 1);
    aLow   := Trunc( aVal);
    aDelta := aVal - aLow;
    Result := aTable[ aLow];
    Result := Result + aDelta * ( aTable[ aLow + 1] - Result);
  end;


  procedure ComputeTables;
  begin
    ComputeTable( UnitsToFrequencyTable          , UnitsToFrequency          );
    ComputeTable( UnitsToFrequencyLinTable       , UnitsToFrequencyLin       );
    ComputeTable( UnitsToFrequencyCubedTable     , UnitsToFrequencyCubed     );
    ComputeTable( UnitsToPhaseDeltaTable         , UnitsToPhaseDelta         );
    ComputeTable( UnitsToPhaseDeltaLinTable      , UnitsToPhaseDeltaLin      );
    ComputeTable( UnitsToSvfFreqTable            , UnitsToSvfFrequency       );
    ComputeTable( UnitsToLfoSlowFrequencyTable   , UnitsToLfoSlowFrequency   );
    ComputeTable( UnitsToLfoSlowPhaseDeltaTable  , UnitsToLfoSlowPhaseDelta  );
    ComputeTable( UnitsToLfoMediumFrequencyTable , UnitsToLfoMediumFrequency );
    ComputeTable( UnitsToLfoMediumPhaseDeltaTable, UnitsToLfoMediumPhaseDelta);
    ComputeTable( UnitsToLfoFastFrequencyTable   , UnitsToLfoFastFrequency   );
    ComputeTable( UnitsToLfoFastPhaseDeltaTable  , UnitsToLfoFastPhaseDelta  );
    ComputeTable( UnitsToLfoBPMFrequencyTable    , UnitsToLfoBPMFrequency    );
    ComputeTable( UnitsToLfoBPMPhaseDeltaTable   , UnitsToLfoBPMPhaseDelta   );
    ComputeTable( UnitsToLfoLinFrequencyTable    , UnitsToLfoLinFrequency    );
    ComputeTable( UnitsToLfoLinPhaseDeltaTable   , UnitsToLfoLinPhaseDelta   );
    ComputeTable( UnitsToSineTable               , UnitsToSine               );
    ComputeTable( UnitsToCosineTable             , UnitsToCosine             );
    ComputeTable( UnitsToScaledEnvTimeTable      , UnitsToScaledEnvTime      );
    ComputeTable( UnitsToScaledSEnvTimeTable     , UnitsToScaledSEnvTime     );
    ComputeTable( CentsToMultiplierTable         , CentsToMultiplierScaled   );
  end;


  function  LookupFrequency          ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyTable          ); end;
  function  LookupFrequencyLin       ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyLinTable       ); end;
  function  LookupFrequencyCubed     ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyCubedTable     ); end;
  function  LookupPhaseDelta         ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToPhaseDeltaTable         ); end;
  function  LookupPhaseDeltaLin      ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToPhaseDeltaLinTable      ); end;
  function  LookupSvfFrequency       ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToSvfFreqTable            ); end;
  function  LookupLfoSlowFrequency   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoSlowFrequencyTable   ); end;
  function  LookupLfoSlowPhaseDelta  ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoSlowPhaseDeltaTable  ); end;
  function  LookupLfoMediumFrequency ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoMediumFrequencyTable ); end;
  function  LookupLfoMediumPhaseDelta( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoMediumPhaseDeltaTable); end;
  function  LookupLfoFastFrequency   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoFastFrequencyTable   ); end;
  function  LookupLfoFastPhaseDelta  ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoFastPhaseDeltaTable  ); end;
  function  LookupLfoBPMFrequency    ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoBPMFrequencyTable    ); end;
  function  LookupLfoBPMPhaseDelta   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoBPMPhaseDeltaTable   ); end;
  function  LookupLfoLinFrequency    ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoLinFrequencyTable    ); end;
  function  LookupLfoLinPhaseDelta   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoLinPhaseDeltaTable   ); end;
  function  LookupSine               ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToSineTable               ); end;
  function  LookupCosine             ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToCosineTable             ); end;
  function  LookupScaledEnvTime      ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToScaledEnvTimeTable      ); end;
  function  LookupScaledSEnvTime     ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToScaledSEnvTimeTable     ); end;


  function  LookupEnvTime( aRange: TEnvRange; aValue: TSignal): TSignal;
  var
    p         : TSignal;
    anOutLow  : TSignal;
    anOutHigh : TSignal;
  begin
    case aRange of
      erSlow   : begin anOutLow := MinEnvTimeSlow  ; anOutHigh := MaxEnvTimeSlow  ; end;
      erMedium : begin anOutLow := MinEnvTimeMedium; anOutHigh := MaxEnvTimeMedium; end;
      else       begin anOutLow := MinEnvTimeFast  ; anOutHigh := MaxEnvTimeFast  ; end;
    end;
    p      := LookupScaledEnvTime( Clip( aValue, 0, 1));
    Result := anOutLow + p * ( anOutHigh - anOutLow);
  end;


  function  LookupCentsMultiplier( const aValue: TSignal): TSignal;
  begin
    Result := TableLookup( Clip( aValue, - CENTS, CENTS) * CENTS_REC, CentsToMultiplierTable);
  end;


  function  MidiFilterToStr( const aValue: TMidiFilter): string;
  var
    i : TMidiSeptet;
  begin
    Result := '{ ';

    for i := Low( TMidiSeptet) to High( TMidiSeptet)
    do begin
      if i in aValue
      then Result := Result + Format( '%d ', [ i], AppLocale);
    end;

    Result := Result + '}';
  end;


  procedure SetSystemRate( aValue: Integer);
  begin
    System_Rate       := aValue;                    // samples / s (Hz)
    System_Rate_2     := System_Rate * 2;           // samples / s (Hz)
    Control_Rate      := System_Rate / CONTROL_DEC; // samples / s (Hz)
    System_Rate_Rec   := 1 / System_Rate;           // s / sample (  s)
    System_Rate_2_Rec := 1 / System_Rate_2;         // s / sample (  s)
  end;


  function  SignalToFractional( aValue: TSignal): Integer; inline;
  begin
    Result := Round( FRACT_SCALING * aValue);
  end;


  function  FractionalToSignal( aValue: Integer): TSignal; inline;
  begin
    Result := FRACT_SCALING_REC * aValue;
  end;


  function  SignalPairArrayToStr( const aValue: TSignalPairArray): string;
  var
    i : Integer;
  begin
    Result := '';

    for i := 0 to Length( aValue) - 1
    do begin
      if Result = ''
      then Result := Format(     '%g, %g', [         aValue[ i].Left, aValue[ i].Right], AppLocale)
      else Result := Format( '%s; %g, %g', [ Result, aValue[ i].Left, aValue[ i].Right], AppLocale);
    end;
  end;


  function  StrToSignalPairArray( const aValue: string; DoFixup: Boolean; aLowX, aHiX: TSignal): TSignalPairArray;

    procedure AddPair( const aData: TSignalPair);
    var
      i : Integer;
      p : Integer;
    begin
      if DoFixup
      then begin
        p := Length( Result);

        for i := 0 to Length( Result) - 2
        do begin
          if ( aData.Left > Result[ i].Left) and ( aData.Left < Result[ i + 1].Left)
          then begin
            p := i + 1;
            Break;
          end;
        end;
        SetLength( Result, Length( Result) + 1);

        for i := Length( Result) - 2 downto p
        do Result[ i + 1] := Result[ i];
      end
      else begin
        SetLength( Result, Length( Result) + 1);
        p := Length( Result) - 1;
      end;

      Result[ p].Left  := aData.Left;
      Result[ p].Right := aData.Right;
    end;

    procedure AddData( anX, anY: TSignal);
    var
      aPair : TSignalPair;
    begin
      aPair.Left  := anX;
      aPair.Right := anY;
      AddPair( aPair);
    end;

    procedure AddStr( const aValue: string);
    var
      aPair : TStringList;
      anX   : TSignal;
      anY   : TSignal;
    begin
      aPair := Explode( aValue, ',');

      try
        if aPair.Count = 2
        then begin
          anX := StrToFloatDef( Trim( aPair[ 0]), Nan);
          anY := StrToFloatDef( Trim( aPair[ 1]), Nan);

          if ( not IsNan( anX)) and ( not IsNan( anY))
          then AddData( anX, anY);
        end;
      finally
        aPair.DisposeOf;
      end;
    end;

  var
    aPairs : TStringList;
    aPair  : string;
  begin
    SetLength( Result, 0);
    aPairs := Explode( aValue, ';');

    if DoFixup
    then begin
      if aPairs.Count = 0
      then aPairs.Add( Format( '%g, 0,0', [ aLowX], AppLocale));

      if aPairs.Count = 1
      then aPairs.Add( Format( '%g, 0,0', [ aHiX], AppLocale));
    end;

    try
      for aPair in aPairs
      do AddStr( Trim( aPair));

      if DoFixup
      then begin
        Result[ 0                  ].Left := aLowX;
        Result[ Length( Result) - 1].Left := aHiX;
      end;
    finally
      aPairs.DisposeOf;
    end;
  end;


  function  InterpolateSignalPairArray( const aData: TSignalPairArray; anX: TSignal): TSignal;
  //
  // This is linear interpolation only, for splined stuff see the BSplines unit.
  //
  // aData must be sorted such that:
  //
  //   for all i in [ 0, Length( aData) - 2] :: aData[ i].Left <= aData[ i + 1].Left
  //
  // or otherwise dX may become <= 0 which can not be handled with the current ccode.

    function FindLarger: Integer;
    var
      i : Integer;
    begin
      Result := Length( aData);

      for i := 0 to Length( aData) - 1
      do begin
        if aData[ i].Left >= anX
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

  var
    p  : Integer;
    dY : TSignal;
    dX : TSignal;
  begin
    p := FindLarger;

    if p < 0                                         // No data in aData
    then Result := 0.0
    else if p = 0                                    // anX < leftmost value in aDaata, use leftmost
    then Result := aData[ 0].Right
    else if p = Length( aData)                       // anX >= rightmost value in aDat, use rightmost
    then Result := aData[ p - 1].Right
    else begin                                       // anX somewhere beteen points p - 1 and p,
                                                     // use linear interpolation in that segment
      dX := aData[ p].Left  - aData[ p - 1].Left;    // Can never be <= 0 ::: !! assuming aData is sorted !!
      dY := aData[ p].Right - aData[ p - 1].Right;

      Result := aData[ p - 1].Right + ( anX - aData[ p - 1].Left ) * ( dY / dX);
    end;
  end;


  function  KnobRangeLin( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else Result := anOutLow + ( aPos / aStepCount) * ( anOutHigh - anOutLow);
  end;


  function  KnobRangeExp( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  var
    p : TSignal;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else begin
      p      := LookupScaledEnvTime( aPos / aStepCount);
      Result := anOutLow + p * ( anOutHigh - anOutLow);
    end;

    if Abs( Result) < MUTE_LIMIT
    then Result := 0;
  end;


type

  TConvUndefined = class( TKnobsNullConverter)
  public
    constructor Create;                                               override;
    function    PosToHint   ( aPos, aSteps: Integer): string;         override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TKnobsBaseConverter = class( TKnobsNullConverter)
  public
    function    PosToHint   ( aPos, aSteps: Integer): string;         override;
  end;


  TConvLinear0_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLinear0_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLambda = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLinearMin1_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLinearMin1_0 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLinearMin4_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvAdd0 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvAdd1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvAdd3 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvOdd = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvQSteps = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdB0_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdB0_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdB0_16 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdB0_32 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdB0_64 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdBKS = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvdBResonator = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvOscFreqCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvOscLinCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvOscCubedCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLfoFreqCoarseFast = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLfoFreqCoarseMedium = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLfoFreqCoarseSlow = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLfoFreqCoarseBPM = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvLfoFreqCoarseLin = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvNoiseColor = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvEnvTimeFast = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvEnvTimeMedium = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvEnvTimeSlow = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvMidiChIn = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvMidiChOut = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvNoteName = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvNoteName12 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTime = class( TKnobsBaseConverter)
  public
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvDelayTimeShort = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvDelayTimeMedium = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvDelayTimeLong = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvFilterFreq = class( TConvOscFreqCoarse)
  public
    constructor Create;                                               override;
  end;


  TConvAvgFilterAlpha = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvCubicClipAlpha = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvSlew = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvCents = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvSemitones = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvSemitonesFull = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvSimple = class( TKnobsBaseConverter)
  public
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TConvMute = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvActivation = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvSolo = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvEnvMute = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvOutType = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvEnvOutType = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvEnvSlopeType = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvWaveShape = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvPulseMode = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TConvPulseDelayMode = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;


  TTalkieFrameRate = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TTalkieSampleRate = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TRoomSize = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;

  end;

  TReverbTime = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TRateMultiplier = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TSwing = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  TSongSpeed = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  THzMeasure = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


{ ========
  TKnobsNullConverter = class
  private
    FSearched     : string;
    FDefaultUnits : string;
  protected
    FName     : string;
    FRate     : TSignalRate;
  public
    property    Name         : string read FName;
    property    DefaultUnits : string read FDefaultUnits;
    property    Searched     : string read FSearched     write FSearched;
  private
}


    function    TKnobsNullConverter.ApplyUnits( const aUnits: string; var aValue: TSignal): Boolean; // virtual;
    const
      OS  = 1.0 / 60.0;
      OTF = 1.0 / 256.0;
      OTS = 1.0 / 360.0;
    var
      Units : string;
    begin
      Result := True;

      if not SameText( DefaultUnits, NO_UNITS)
      then begin
        if aUnits = ''
        then Units := DefaultUnits
        else Units := aUnits;

        if      SameText( Units, 'Hz')
        then aValue := FrequencyToUnits( aValue)
        else if SameText( Units, 'kHz')
        then aValue := FrequencyToUnits( 1000 * aValue)
        else if SameText( Units, 'mHz')
        then aValue := FrequencyToUnits( aValue * 1e-3)
        else if SameText( Units, 'bpm')
        then aValue := FrequencyToUnits( aValue * OS)
        else if SameText( Units, 'degrees') or SameText( aUnits, 'deg')
        then aValue := aValue * OTS
        else if SameText( Units, 'radians') or SameText( aUnits, 'rad')
        then aValue := aValue * TWO_PI_REC
        else if SameText( Units, 'percent') or SameText( aUnits, '%')
        then aValue := aValue * 1e-2
        else if SameText( Units, 'ms')
        then aValue := aValue * 1e-3
        else if SameText( Units, 's')
        then aValue := aValue * 1e-6
        else if SameText( Units, 's')
        then // nothing
        else if SameText( Units, 'nn')
        then aValue := NoteNumberToUnits( aValue)
        else if SameText( Units, 'qsteps') or SameText( aUnits, 'q')
        then aValue := aValue * 1e-3
        else if SameText( Units, 'ct')
        then aValue := aValue * OTF
        else Result := False;
      end;
    end;


    function    TKnobsNullConverter.StringToValue( const aStringValue: string; var aValue: TSignal): Boolean; // virtual;
    begin
      Result := True;

      if IsNoteName( aStringValue)
      then aValue := NoteNameToUnits( aStringValue, UNDEFINED)
      else if IsTimeValue( aStringValue)
      then aValue := 1 / TimeToSeconds( aStringValue, UNDEFINED)
      else if SameText( aStringValue, 'on')
      then aValue := 1.0
      else if SameText( aStringValue, 'off')
      then aValue := 0.0
      else Result := TextToFloat( aStringValue, aValue, AppLocale);

      if not Result
      then aValue := UNDEFINED;
    end;


    function    TKnobsNullConverter.PostProcess( const aValue: TSignal): TSignal; // virtual;
    begin
      Result := aValue;
    end;


//  public

    constructor TKnobsNullConverter.Create; // virtual;
    begin
      FName         := '';
      FSearched     := '';
      FRate         := srAudio;
      FDefaultUnits := NO_UNITS;
    end;


    function    TKnobsNullConverter.PosToHint( aPos, aSteps: Integer): string; // virtual;
    begin
      Result := IntToStr( aPos);
    end;


    function    TKnobsNullConverter.PosToDisplay( aPos, aSteps: Integer): string; // virtual;
    begin
      Result := IntToStr( aPos);
    end;


    function    TKnobsNullConverter.PosToValue( aPos, aSteps: Integer): TSignal; // virtual;
    begin
      Result := aPos;
    end;


    function    TKnobsNullConverter.DisplayToValue( const aDisplay: string; var aValue: TSignal): Boolean;
    var
      aParts : TStringList;
    begin
      aValue := UNDEFINED;
      Result := False;
      aParts := Explode( aDisplay, ' ');

      try
        if aParts.Count > 0
        then Result := StringToValue( Trim( aParts[ 0]), aValue);

        if Result and ( aParts.Count > 1)
        then Result := ApplyUnits( Trim( aParts[ 1]), aValue);

        if Result
        then aValue := PostProcess( aValue);
      finally
        aParts.DisposeOf;
      end;
    end;


{ ========
  TKnobsConverterUndefined = class( TKnobsNullConverter)
  public
}

    constructor TConvUndefined.Create; // override;
    begin
      FName := 'undefined converter';
      FRate := srAudio;
    end;


    function    TConvUndefined.PosToHint( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'undefined %s'^M'does %s'^M'%f', [ Searched, inherited, PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvUndefined.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'undefined %s does %s [%f]', [ Searched, inherited, PosToValue( aPos, aSteps)], AppLocale);
    end;


{ ========
  TKnobsBaseConverter = class( TKnobsNullConverter)
  public
}

    function    TKnobsBaseConverter.PosToHint( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( ' value: %s '^M' position: %d ', [ PosToDisplay( aPos, aSteps), aPos], AppLocale);
    end;


{ ========
  TConvLinear0_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinear0_1.Create; // override;
    begin
      FName := 'Linear[0,1]';
      FRate := srAudio;
    end;


    function    TConvLinear0_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvLinear0_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvLinear0_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinear0_4.Create; // override;
    begin
      FName := 'Linear[0,4]';
      FRate := srAudio;
    end;


    function    TConvLinear0_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvLinear0_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 4);
    end;


{ ========
  TConvLambda = class( TKnobsBaseConverter)
  public
}

    constructor TConvLambda.Create; // override;
    begin
      FName := 'Lambda';
      FRate := srAudio;
    end;


    function    TConvLambda.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvLambda.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 5, 1);
    end;


{ ========
  TConvLineMin1_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin1_1.Create; // override;
    begin
      FName := 'Linear[-1,1]';
      FRate := srAudio;
    end;


    function    TConvLinearMin1_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvLinearMin1_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1, 1);
    end;



{ ========
  TConvLineMin1_0 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin1_0.Create; // override;
    begin
      FName := 'Linear[-1,0]';
      FRate := srAudio;
    end;


    function    TConvLinearMin1_0.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvLinearMin1_0.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1, 0);
    end;


{ ========
  TConvLineMin4_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin4_4.Create; // override;
    begin
      FName := 'Linear[-4,4]';
      FRate := srAudio;
    end;


    function    TConvLinearMin4_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvLinearMin4_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -4, 4);
    end;


{ ========
  TConvAdd0 = class( TKnobsBaseConverter)
  public
}

    constructor TConvAdd0.Create; // override;
    begin
      FName := 'Add0';
      FRate := srAudio;
    end;


    function    TConvAdd0.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvAdd0.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TConvAdd1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvAdd1.Create; // override;
    begin
      FName := 'Add1';
      FRate := srAudio;
    end;


    function    TConvAdd1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvAdd1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvAdd3 = class( TKnobsBaseConverter)
  public
}

    constructor TConvAdd3.Create; // override;
    begin
      FName := 'Add3';
      FRate := srAudio;
    end;


    function    TConvAdd3.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvAdd3.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 3;
    end;


{ ========
  TConvOdd = class( TKnobsBaseConverter)
  public
}

    constructor TConvOdd.Create; // override;
    begin
      FName := 'Odd';
      FRate := srAudio;
    end;


    function    TConvOdd.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := IntToStr( 2 * aPos + 1);
    end;


    function    TConvOdd.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := 2 * aPos + 1;
    end;


{ ========
  TConvQSteps = class( TKnobsBaseConverter)
  public
}

    constructor TConvQSteps.Create; // override;
    begin
      FName := 'QSteps';
      FRate := srAudio;
    end;


    function    TConvQSteps.PosToDisplay( aPos, aSteps: Integer): string; // override;
    const
      OSF = 1.0 / 64.0;
    var
      R : TRational;
    begin
      R := PosToValue( aPos, aSteps) * OSF;
      R.Invert;
      Result := Format( '%s', [ R.AsString], AppLocale);
    end;


    function    TConvQSteps.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvdB0_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_1.Create; // override;
    begin
      FName         := 'dB[0,1]';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdB0_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;


    function    TConvdB0_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvdB0_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_4.Create; // override;
    begin
      FName         := 'dB[0,4]';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdB0_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;


    function    TConvdB0_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 4);
    end;


{ ========
  TConvdB0_16 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_16.Create; // override;
    begin
      FName         := 'dB[0,16]';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdB0_16.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;


    function    TConvdB0_16.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 16);
    end;


{ ========
  TConvdB0_32 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_32.Create; // override;
    begin
      FName         := 'dB[0,32]';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdB0_32.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;


    function    TConvdB0_32.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 32);
    end;


{ ========
  TConvdB0_64 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_64.Create; // override;
    begin
      FName         := 'dB[0,64]';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdB0_64.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;


    function    TConvdB0_64.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1600);
    end;


{ ========
  TConvdBKS = class( TKnobsBaseConverter)
  public
}

    constructor TConvdBKS.Create; // override;
    begin
      FName         := 'dB KS';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdBKS.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.2f dB', [ aValue], AppLocale);
    end;


    function    TConvdBKS.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.8, 1);
    end;



{ ========
  TConvdbResonator = class( TKnobsBaseConverter)
  public
}

    constructor TConvdbResonator.Create; // override;
    begin
      FName         := 'dB Resonator';
      FDefaultUnits := 'dB';
      FRate         := srAudio;
    end;


    function    TConvdbResonator.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := UnitsTodB( aValue);

      if aValue <= - MAX_DB_VAL
      then Result := 'off'
      else Result := Format( '%.6f dB', [ aValue], AppLocale);
    end;


    function    TConvdbResonator.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.999, 0.999999999999999);
    end;



{ ========
  TConvOscFreqCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscFreqCoarse.Create; // override;
    begin
      FName         := 'OscFreqCoarse';
      FDefaultUnits := 'Hz';
      FRate         := srAudio;
    end;


    function    TConvOscFreqCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequency( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvOscFreqCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvOscLinCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscLinCoarse.Create; // override;
    begin
      FName         := 'OscLinCoarse';
      FDefaultUnits := 'Hz';
      FRate         := srAudio;
    end;


    function    TConvOscLinCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequencyLin( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvOscLinCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvOscCubedCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscCubedCoarse.Create; // override;
    begin
      FName         := 'OscCubedCoarse';
      FDefaultUnits := 'Hz';
      FRate         := srAudio;
    end;


    function    TConvOscCubedCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequencyCubed( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvOscCubedCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseFast = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseFast.Create; // override;
    begin
      FName         := 'LfoFreqCoarseFast';
      FDefaultUnits := 'Hz';
      FRate         := srControl;
    end;


    function    TConvLfoFreqCoarseFast.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoFastFrequency( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvLfoFreqCoarseFast.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseMedium.Create; // override;
    begin
      FName         := 'LfoFreqCoarseMedium';
      FDefaultUnits := 'Hz';
      FRate         := srControl;
    end;


    function    TConvLfoFreqCoarseMedium.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoMediumFrequency( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvLfoFreqCoarseMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseSlow = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseSlow.Create; // override;
    begin
      FName         := 'LfoFreqCoarseSlow';
      FDefaultUnits := 'Hz';
      FRate         := srControl;
    end;


    function    TConvLfoFreqCoarseSlow.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoSlowFrequency( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvLfoFreqCoarseSlow.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvLfoFreqCoarseBPM = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseBPM.Create; // override;
    begin
      FName         := 'LfoFreqCoarseBPM';
      FDefaultUnits := 'bpm';
      FRate         := srControl;
    end;


    function    TConvLfoFreqCoarseBPM.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoBPMFrequency( aValue);
      Result := FreqToBPMStr( aValue);
    end;


    function    TConvLfoFreqCoarseBPM.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvLfoFreqCoarseLin = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseLin.Create; // override;
    begin
      FName         := 'LfoFreqCoarseLin';
      FDefaultUnits := 'Hz';
      FRate         := srControl;
    end;


    function    TConvLfoFreqCoarseLin.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoLinFrequency( aValue);
      Result := FreqToStr( aValue);
    end;


    function    TConvLfoFreqCoarseLin.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvNoiseColor = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoiseColor.Create; // override;
    begin
      FName         := 'NoiseColor';
      FDefaultUnits := '';
      FRate         := srAudio;
    end;


    function    TConvNoiseColor.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      if aPos = 0
      then Result := 'off'
      else if aPos = aSteps
      then Result := 'white'
      else Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvNoiseColor.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvEnvTimeFast = class( TKnobsBaseConverter)
  public
}

    constructor TConvEnvTimeFast.Create; // override;
    begin
      FName         := 'EnvTimeFast';
      FDefaultUnits := 's';
      FRate         := srAudio;
    end;


    function    TConvEnvTimeFast.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      if aSteps <= 1
      then aValue := MinEnvTimeFast
      else aValue := LookupEnvTime( erFast, aPos / ( aSteps - 1));

      Result := TimeToStr( aValue);
    end;


    function    TConvEnvTimeFast.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvEnvTimeMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvEnvTimeMedium.Create; // override;
    begin
      FName         := 'EnvTimeMedium';
      FDefaultUnits := 's';
      FRate         := srAudio;
    end;


    function    TConvEnvTimeMedium.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      if aSteps <= 1
      then aValue := MinEnvTimeMedium
      else aValue := LookupEnvTime( erMedium, aPos / ( aSteps - 1));

      Result := TimeToStr( aValue);
    end;


    function    TConvEnvTimeMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvEnvTimeSlow = class( TKnobsBaseConverter)
  public
}

    constructor TConvEnvTimeSlow.Create; // override;
    begin
      FName         := 'EnvTimeSlow';
      FDefaultUnits := 's';
      FRate         := srAudio;
    end;


    function    TConvEnvTimeSlow.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      if aSteps <= 1
      then aValue := MinEnvTimeSlow
      else aValue := LookupEnvTime( erSlow, aPos / ( aSteps - 1));

      Result := TimeToStr( aValue);
    end;


    function    TConvEnvTimeSlow.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvMidiCCIn = class( TKnobsBaseConverter)
  public
}

    constructor TConvMidiChIn.Create; // override;
    begin
      FName         := 'MidiChIn';
      FDefaultUnits := NO_UNITS;
      FRate         := srAudio;
    end;


    function    TConvMidiChIn.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      if  aPos = aSteps - 1
      then Result := 'omni'
      else Result := Format( 'ch %.0f', [ PosToValue( aPos, aSteps) + 1], AppLocale);
    end;


    function    TConvMidiChIn.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TConvMidiCCOut = class( TKnobsbaseConverter)
  public
}

    constructor TConvMidiChOut.Create; // override;
    begin
      FName         := 'MidiChOut';
      FDefaultUnits := NO_UNITS;
      FRate         := srAudio;
    end;


    function    TConvMidiChOut.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'ch %.0f', [ PosToValue( aPos, aSteps) + 1], AppLocale);
    end;


    function    TConvMidiChOut.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;



{ ========
  TConvNoteName = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoteName.Create; // override;
    begin
      FName         := 'NoteName';
      FDefaultUnits := '';
      FRate         := srAudio;
    end;


    function    TConvNoteName.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := UnitsToNoteName( PosToValue( aPos, aSteps));
    end;


    function    TConvNoteName.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvNoteName12 = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoteName12.Create; // override;
    begin
      FName         := 'NoteName12';
      FDefaultUnits := '';
      FRate         := srAudio;
    end;


    function    TConvNoteName12.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := UnitsToNoteName12( PosToValue( aPos, aSteps));
    end;


    function    TConvNoteName12.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvDelayTime = class( TKnobsBaseConverter)
  public
}

    function    TConvDelayTime.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := TimeToStr( PosToValue( aPos, aSteps));
    end;


{ ========
  TConvDelayTimeShort = class( TConvDelayTime)
  public
}

    constructor TConvDelayTimeShort.Create; // override;
    begin
      FName         := 'DelayTimeShort';
      FDefaultUnits := 's';
      FRate         := srAudio;
    end;


    function    TConvDelayTimeShort.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 100e-6, 3e-2);
    end;


{ ========
  TConvDelayTimeMedium = class( TConvDelayTime)
  public
}

    constructor TConvDelayTimeMedium.Create; // override;
    begin
      FName         := 'DelayTimeMedium';
      FDefaultUnits := 's';
      FRate         := srAudio;
    end;


    function    TConvDelayTimeMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 10e-3, 3);
    end;


{ ========
  TConvDelayTimeLong = class( TConvDelayTime)
  public
}

    constructor TConvDelayTimeLong.Create; // override;
    begin
      FName         := 'DelayTimeLong';
      FDefaultUnits := 's';
      FRate         := srAudio;
    end;


    function    TConvDelayTimeLong.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1, 300);
    end;


{ ========
  TConvFilterFreq = class( TConvOscFreqCoarse)
  public
}

    constructor TConvFilterFreq.Create; // override;
    begin
      FName         := 'FilterFreq';
      FDefaultUnits := 'Hz';
      FRate         := srAudio;
    end;


{ ========
  TConvAvgFilterAlpha = class( TKnobsBaseConverter)
  public
}

    constructor TConvAvgFilterAlpha.Create; // override;
    begin
      FName         := 'FilterAvgAlpha';
      FDefaultUnits := '';
      FRate         := srAudio;
    end;


    function    TConvAvgFilterAlpha.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.5f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvAvgFilterAlpha.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvCubicClipAlpha = class( TKnobsBaseConverter)
  public
}

    constructor TConvCubicClipAlpha.Create; // override;
    begin
      FName         := 'CubicClipAlpha';
      FDefaultUnits := NO_UNITS;
      FRate         := srAudio;
    end;


    function    TConvCubicClipAlpha.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvCubicClipAlpha.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps, 0.1, 10);
    end;


{ ========
  TConvSlew = class( TKnobsBaseConverter)
  public
}

    constructor TConvSlew.Create; // override;
    begin
      FName         := 'Slew';
      FDefaultUnits := NO_UNITS;
      FRate         := srAudio;
    end;


    function    TConvSlew.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.5f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvSlew.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvCents = class( TKnobsBaseConverter)
  public
}

    constructor TConvCents.Create; // override;
    begin
      FName         := 'Cents';
      FDefaultUnits := 'ct';
      FRate         := srAudio;
    end;


    function    TConvCents.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f ct', [ 50 * 256 * PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvCents.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    const
      OTSF  = 1.0 / 256.0;             // Quarter not positive
      OTSFN = - OTSF;                  // Quarter note negative
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, OTSFN, OTSF);
    end;


{ ========
  TConvSemitones = class( TKnobsBaseConverter)
  public
}

    constructor TConvSemitones.Create; // override;
    begin
      FName         := 'Semitones';
      FDefaultUnits := NO_UNITS;
      FRate         := srAudio;
    end;


    function    TConvSemitones.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ NOTE_SCALING * PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TConvSemitones.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    const
      SCALE = 24.0 * NOTE_SCALING_REC;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, - SCALE, SCALE);
    end;


{ ========
  TConvSemitonesFull = class( TKnobsBaseConverter)
  public
}

    constructor TConvSemitonesFull.Create; // override;
    begin
      FName         := 'SemitonesFull';
      FDefaultUnits := NO_UNITS;
      FRate         := srAudio;
    end;


    function    TConvSemitonesFull.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue   : TSignal;
      aSign    : Integer;
      anIntVal : Integer;
      aRepr    : string;
    begin
      aValue   := NOTE_SCALING * PosToValue( aPos, aSteps);
      aSign    := Sign( aValue);
      anIntVal := Round( Abs( aValue));
      aRepr    := Format( '%d semitones', [ anIntVal], AppLocale);

      if aSign < 0
      then Result := Format( '-%s', [ aRepr], AppLocale)
      else Result := aRepr;
    end;


    function    TConvSemitonesFull.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1, 1);
    end;


{ ========
  TConvSimple = class( TKnobsBaseConverter)
  public
}

    function    TConvSimple.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TConvMute = class( TConvSimple)
  public
}

    constructor TConvMute.Create; // override;
    begin
      FName         := 'mute';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvMute.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'active';
        1 :  Result := 'muted';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvActivation = class( TConvSimple)
  public
}

    constructor TConvActivation.Create; // override;
    begin
      FName         := 'activation';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvActivation.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'off';
        1 :  Result := 'active';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvSolo = class( TConvSimple)
  public
}

    constructor TConvSolo.Create; // override;
    begin
      FName         := 'solo';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvSolo.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'normal';
        1 :  Result := 'solo';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvEnvMute = class( TConvSimple)
  public
}

    constructor TConvEnvMute.Create; // override;
    begin
      FName         := 'envmute';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvEnvMute.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'active';
        1 :  Result := 'muted';
        2 :  Result := 'pass through';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvOutType = class( TConvSimple)
  public
}

    constructor TConvOutType.Create; // override;
    begin
      FName         := 'outtype';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvOutType.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Normal';
        1 :  Result := 'Inverted';
        2 :  Result := 'Positive';
        3 :  Result := 'Positve Inverted';
        4 :  Result := 'Negative';
        5 :  Result := 'Negative Inverted';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvEnvOutType = class( TConvSimple)
  public
}

    constructor TConvEnvOutType.Create; // override;
    begin
      FName         := 'envouttype';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvEnvOutType.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Positive';
        1 :  Result := 'Positve Inverted';
        2 :  Result := 'Negative';
        3 :  Result := 'Negative Inverted';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvEnvSlopeType = class( TConvSimple)
  public
}

    constructor TConvEnvSlopeType.Create; // override;
    begin
      FName         := 'envslopetype';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvEnvSlopeType.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'log';
        1 :  Result := 'lin';
        2 :  Result := 'exp';
        3 :  Result := 's';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvWaveShape = class( TConvSimple)
  public
}

    constructor TConvWaveShape.Create; // override;
    begin
      FName         := 'waveshape';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvWaveShape.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Sine';
        1 :  Result := 'Tri';
        2 :  Result := 'Saw';
        3 :  Result := 'Square';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvPulseMode = class( TConvSimple)
  public
}

    constructor TConvPulseMode.Create; // override;
    begin
      FName         := 'pulsemode';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvPulseMode.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Square';
        1 :  Result := 'Pulse';
        2 :  Result := 'Input';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvPulseDelayMode = class( TConvSimple)
  public
}

    constructor TConvPulseDelayMode.Create; // override;
    begin
      FName         := 'pulse delay mode';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TConvPulseDelayMode.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Delay';
        1 :  Result := 'Stretch';
        else Result := 'unknown';
      end;
    end;


{ ========
  TTalkieFrameRate = class( TKnobsBaseConverter)
  public
}

    constructor TTalkieFrameRate.Create; // override;
    begin
      FName         := 'talkie frame rate';
      FDefaultUnits := 'Hz';
      FRate         := srControl;
    end;


    function    TTalkieFrameRate.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f Hz', [ 80.0 * PosToValue( aPos, aSteps)]);
    end;


    function    TTalkieFrameRate.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TTalkieSampleRate = class( TKnobsBaseConverter)
  public
}

    constructor TTalkieSampleRate.Create; // override;
    begin
      FName         := 'talkie sample rate';
      FDefaultUnits := 'Hz';
      FRate         := srControl;
    end;


    function    TTalkieSampleRate.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.1f Hz', [ 16000.0 * PosToValue( aPos, aSteps)]);
    end;


    function    TTalkieSampleRate.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TRoomSize = class( TKnobsBaseConverter)
  public
}

    constructor TRoomSize.Create; // override;
    begin
      FName         := 'room size';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TRoomSize.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f m', [ PosToValue( aPos, aSteps)]);
    end;


    function    TRoomSize.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1, 6400);
    end;


{ ========
  TReverbTime = class( TKnobsBaseConverter)
  public
}

    constructor TReverbTime.Create; // override;
    begin
      FName         := 'reverb time';
      FDefaultUnits := 's';
      FRate         := srControl;
    end;


    function    TReverbTime.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := TimeToStr( PosToValue( aPos, aSteps));
    end;


    function    TReverbTime.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.050, 1800);
    end;


{ ========
  TRateMultiplier = class( TKnobsBaseConverter)
  public
}

    constructor TRateMultiplier.Create; // override;
    begin
      FName         := 'rate multiplier';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TRateMultiplier.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := RateMultiplierToStr( PosToValue( aPos, aSteps));
    end;


    function    TRateMultiplier.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, - 23, 24);
    end;


{ ========
  TSwing = class( TKnobsBaseConverter)
  public
}

    constructor TSwing.Create; // override;
    begin
      FName         := 'swing';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TSwing.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.1f %%', [ 100.0 * PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TSwing.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0.50, 0.96875);
    end;



{ ========
  TSongSpeed = class( TKnobsBaseConverter)
  public
}

    constructor TSongSpeed.Create; // override;
    begin
      FName         := 'song speed';
      FDefaultUnits := NO_UNITS;
      FRate         := srControl;
    end;


    function    TSongSpeed.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;


    function    TSongSpeed.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0.1, 5.0);
    end;


{ ========
  THzMeasure = class( TKnobsBaseConverter)
  public
}

    constructor THzMeasure.Create; // override;
    begin
      FName         := 'HzMeasure';
      FDefaultUnits := 's';
      FRate         := srControl;
    end;


    function    THzMeasure.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        1 :  Result := '100 Hz';
        2 :  Result := '10 Hz';
        3 :  Result := '1 Hz';
        4 :  Result := '10 s';
        else Result := '1 kHz';
      end;
    end;


    function    THzMeasure.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TKnobsConverters = class( TStringList)
  private
 }

    function    TKnobsConverters.Converter( const anIndex: string): TKnobsNullConverter;
    var
      p : Integer;
    begin
      p  := IndexOf( anIndex);

      if p < 0
      then begin
        p  := IndexOf( 'undefined converter');

        if p >= 0
        then (Objects[ p] as TKnobsNullConverter).Searched := anIndex
        else raise ENoDefaultConverter.Create( 'TKnobsConverters should at the very least have a default converter ("undefined converter", TKnobsNullConverter).');
      end;

      Result := Objects[ p] as TKnobsNullConverter;
    end;


 // public

    constructor TKnobsConverters.Create;
    begin
      inherited;
      OwnsObjects   := True;
      Duplicates    := dupError;
      Sorted        := True;
      CaseSensitive := False;
      RegisterConverters(
        [
          TKnobsNullConverter      , TConvUndefined           , TConvLinear0_1           , TConvLinear0_4           ,
          TConvLambda              , TConvLinearMin1_1        , TConvLinearMin1_0        , TConvLinearMin4_4        ,
          TConvdB0_1               , TConvdB0_4               , TConvdB0_16              , TConvdB0_32              ,
          TConvdB0_64              , TConvdBKS                , TConvdBResonator         , TConvOscFreqCoarse       ,
          TConvNoiseColor          , TConvOscLinCoarse        , TConvOscCubedCoarse      , TConvLfoFreqCoarseFast   ,
          TConvLfoFreqCoarseMedium , TConvLfoFreqCoarseSlow   , TConvLfoFreqCoarseBPM    , TConvLfoFreqCoarseLin    ,
          TConvEnvTimeFast         , TConvEnvTimeMedium       , TConvEnvTimeSlow         , TConvAdd0                ,
          TConvAdd1                , TConvAdd3                , TConvOdd                 , TConvQSteps              ,
          TConvMidiChIn            , TConvMidiChOut           , TConvNoteName            , TConvNoteName12          ,
          TConvDelayTimeShort      , TConvDelayTimeMedium     , TConvDelayTimeLong       , TConvFilterFreq          ,
          TConvAvgFilterAlpha      , TConvCubicClipAlpha      , TConvSlew                , TConvCents               ,
          TConvSemitones           , TConvSemitonesFull       , TConvMute                , TConvActivation          ,
          TConvSolo                , TConvEnvMute             , TConvOutType             , TConvEnvOutType          ,
          TConvEnvSlopeType        , TConvWaveShape           , TConvPulseMode           , TConvPulseDelayMode      ,
          TTalkieFrameRate         , TTalkieSampleRate        , TRoomSize                , TReverbTime              ,
          TRateMultiplier          , TSwing                   , TSongSpeed               , THzMeasure
        ]
      );
    end;


    procedure   TKnobsConverters.RegisterConverter( aClass: TKnobsConverterClass);
    var
      C: TKnobsNullConverter;
    begin
      C := aClass.Create;
      AddObject( C.Name, C);
    end;


    procedure   TKnobsConverters.RegisterConverters( const aClasses: array of TKnobsConverterClass);
    var
      C: TKnobsConverterClass;
    begin
      for C in aClasses
      do RegisterConverter( C);
    end;


    function    TKnobsConverters.PosToHint( const aName: string; aPos, aSteps: Integer): string;
    begin
      Result := Converter( aName).PosToHint( aPos, aSteps);
    end;


    function    TKnobsConverters.PosToDisplay( const aName: string; aPos, aSteps: Integer): string;
    begin
      Result := Converter( aName).PosToDisplay( aPos, aSteps);
    end;


    function    TKnobsConverters.PosToValue( const aName: string; aPos, aSteps: Integer): TSignal;
    begin
      Result := Converter( aName).PosToValue( aPos, aSteps);
    end;


    function    TKnobsConverters.DisplayToValue( const aName, aDisplay: string; var aValue: TSignal): Boolean;
    begin
      Result := Converter( aName).DisplayToValue( aDisplay, aValue);
    end;


  // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure InitializeUnit;
  begin
    ReferenceA        := 440.0;
    NotesPerOctave    := 12.0;
    NotesPerOctaveRec := 1.0 / NotesPerOctave;
    MiddleNote        := 69;
    SetSystemRate( 44100);
    ComputeTables;
  end;


  procedure CreateConverters;
  begin
    LocConverters := TKnobsConverters.Create;
  end;


  procedure FreeConverters;
  begin
    FreeAndNil( LocConverters);
  end;


initialization

  CreateConverters;
  InitializeUnit;

finalization

  FreeConverters;

end.

